
// $Id: BddManager.java,v 1.92 2003/02/01 00:20:04 fenyo Exp $

import java.net.*;
import java.io.*;
import java.util.*;
import java.lang.*;
import java.lang.reflect.*;

import java.text.*;

import java.nio.*;
import java.nio.channels.*;

import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;

import org.w3c.dom.*;

import java.rmi.*;
import java.rmi.server.*;

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.awt.event.*;
import java.awt.font.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.text.*;
import javax.swing.tree.*;
import javax.swing.table.*;
import javax.swing.event.*;

import java.beans.*;

///////////////////////////////////////////////////////////

class MyException extends Exception {
    public MyException(String message) {
	super(message);
    }
}

///////////////////////////////////////////////////////////

class Trace implements Cloneable {
    // access by any thread
    public Trace() {
	Debug.debugln(this, "constructor call");
    }

    protected void finalize() throws Throwable {
	Debug.debugln(this, "finalizer call");
	super.finalize();
    }

    public Object clone() throws CloneNotSupportedException {
	return super.clone();
    }
}

///////////////////////////////////////////////////////////

class Debug {
    // access by any thread
    public static void debugln(String s) {
	if (Conf.isDebug() == true) System.out.println(s);
    }

    // access by any thread
    public static void errorln(String s) {
	System.out.println("ERROR: " + s);
    }

    // access by any thread
    public static void debugln(Object o, String s) {
	if (Conf.isDebug() == true)
	    System.out.println("[thr:" + Thread.currentThread().getName() +
			       "|obj:" + o.toString() + "]: " + s);
    }

    // access by any thread
    public static void errorln(Object o, String s) {
	System.out.println("ERROR: [thr:" + Thread.currentThread().getName() +
			   "|obj:" + o.toString() + "]: " + s);
    }

    // access by any thread
    public static void logException(Object o, Exception e) {
	System.out.println("[thr:" + Thread.currentThread().getName() +
			   "|obj:" + o.toString() + "]: exception thrown:\r\n  " +
                           e.getClass().toString() + " [" +
                           e.getMessage() + "]");
    }

    // access by any thread
    public static void logException(Object o, Exception e, String s) {
	System.out.println("[thr:" + Thread.currentThread().getName() +
			   "|obj:" + o.toString() + "]: exception thrown:\r\n  " +
                           e.getClass().toString() + " [" +
                           e.getMessage() + "] " + s);
    }

    // access by any thread
    public static void logExceptionAndExit(Object o, Exception e) {
	logException(o, e);
	System.exit(1);
    }

    // access by any thread
    public static void logExceptionAndExit(Object o, Exception e, String s) {
	logException(o, e, s);
	System.exit(1);
    }

    // access by any thread
    public static void logException(Exception e) {
	logException(e, e);
    }

    // access by any thread
    public static void logException(Exception e, String s) {
	logException(e, e, s);
    }

    // access by any thread
    public static void logExceptionAndExit(Exception e) {
	logException(e);
	System.exit(1);
    }

    // access by any thread
    public static void logExceptionAndExit(Exception e, String s) {
	logException(e, s);
	System.exit(1);
    }
}

////////////////////////////////////////////////////////////

class DebugGUI {
    public static JFrame frame = null;
    public static View view = null;
    public static JScrollPane scrollPane = null;
    public static VPict vpict = null;

    public static final Random random = new Random();
}

class DebugBdd {
    public static RemoteAccess remoteAccess = null;
}

////////////////////////////////////////////////////////////

class Conf {
    private static boolean debug = false;
    private static int clientsPort = 8008;
    private static int clientsBacklog = 15;
    private static String netTemplatesDir = ".";
    private static String logDir = ".";
    private static String rmiRegistry = "//localhost:1099";
    private static String mode = "both";

    // access by main thread
    public static void setParameters() {
 	if (System.getProperty("gnetwatch.debug") != null)
 	    debug = Boolean.getBoolean("gnetwatch.debug");

	if (System.getProperty("gnetwatch.port") != null)
	    clientsPort = Integer.parseInt(System.getProperty("gnetwatch.port"));

	if (System.getProperty("gnetwatch.tmpldir") != null)
	    netTemplatesDir = System.getProperty("gnetwatch.tmpldir");

	if (System.getProperty("gnetwatch.logdir") != null)
	    logDir = System.getProperty("gnetwatch.logdir");

	if (System.getProperty("gnetwatch.registry") != null)
	    rmiRegistry = System.getProperty("gnetwatch.registry");

	if (System.getProperty("gnetwatch.mode") != null &&
	    System.getProperty("gnetwatch.mode").equals("server") == false &&
	    System.getProperty("gnetwatch.mode").equals("client") == false &&
	    System.getProperty("gnetwatch.mode").equals("both") == false) {
	    Debug.errorln("invalid mode");
	    System.exit(1);
	}
	if (System.getProperty("gnetwatch.mode") != null)
	    mode = System.getProperty("gnetwatch.mode");
    }

    // access by any thread
    public static boolean isDebug() {
	return debug;
    }

    // access by any thread
    public static int getClientsPort() {
	return clientsPort;
    }

    // access by any thread
    public static int getClientsBacklog() {
	return clientsBacklog;
    }

    // access by any thread
    public static String getNetTemplatesDir() {
	return netTemplatesDir;
    }

    // access by any thread
    public static String getLogDir() {
	return logDir;
    }

    // access by any thread
    public static String getRmiRegistry() {
	return rmiRegistry;
    }

    // access by any thread
    public static String getMode() {
	return mode;
    }
}

////////////////////////////////////////////////////////////

// EntryKey implements comparable to allow TreeMap to support EntryKey as keys
class EntryKey implements Serializable, Comparable {
    private final String identity, key; // constant strings -> no need to synchronize

    public EntryKey(String identity, String key) {
	this.identity = identity;
	this.key = key;
    }

    public int compareTo(Object o) {
	int comp = ((EntryKey) o).getIdentity().compareTo(identity);
	return comp != 0 ? comp : ((EntryKey) o).getKey().compareTo(key);
    }

    public int hashCode() {
	return identity.hashCode() ^ key.hashCode();
    }

    public boolean equals(Object obj) {
	if (obj != null && obj instanceof EntryKey &&
	    ((EntryKey) obj).identity.equals(identity) &&
 	    ((EntryKey) obj).key.equals(key)) return true;
 	return false;
    }

    public String getIdentity() {
	return identity;
    }

    public String getKey() {
	return key;
    }

    public String dump() {
	return new String("[" + identity + "," + key + "]");
    }
}

class EntryVal implements Serializable {
    private final String value, reference; // constant strings -> no need to synchronize
    private final Date timestamp; // constant date -> no need to synchronize

    public EntryVal(String value, String reference, Date timestamp) {
	this.value = value;
	this.reference = reference;
	this.timestamp = timestamp;
    }

    public int hashCode() {
	return value.hashCode() ^ reference.hashCode() ^ timestamp.hashCode();
    }

    public boolean equals(Object obj) {
	if (obj != null && obj instanceof EntryVal &&
	    ((EntryVal) obj).value.equals(value) &&
 	    ((EntryVal) obj).reference.equals(reference) &&
	    ((EntryVal) obj).timestamp.equals(timestamp)) return true;
 	return false;
    }

    public String getValue() {
	return value;
    }

    public String getReference() {
	return reference;
    }

    public Date getTimestamp() {
	return timestamp;
    }

    public String dump() {
	return new String("[" + value + "," + reference + "," + timestamp.toString() + "]");
    }
}

////////////////////////////////////////////////////////////

// State uses only MapTree instead of HashMap to increase performances
// since it is often serialized across the network
class State extends Trace implements Serializable, Cloneable {
    // State instances are used as a lock to embedded data

    private boolean updated = false;
    private Integer updated_lock = new Integer(0);

    public Date date = new Date(Long.MIN_VALUE);
    public TreeMap map = new TreeMap();

    public boolean waitForUpdate() throws InterruptedException {
	boolean retval = true;

	synchronized (updated_lock) {
	    if (updated == false) updated_lock.wait(2000);
	    retval = updated;
	    updated = false;
	}

	return retval;
    }

    public void update() {
	synchronized (updated_lock) {
	    updated = true;
	    updated_lock.notify();
	}
    }

    public void wakeup() {
	updated_lock.notify();
    }

    public synchronized void put(EntryKey infoKey, EntryVal infoVal) {
	if (map.containsKey(infoKey) == false)  map.put(infoKey, new TreeMap());
	TreeMap reflist = (TreeMap) map.get(infoKey);
	reflist.put(infoVal.getReference(), infoVal);

	update();
    }

    public synchronized void remove(EntryKey infoKey, EntryVal infoVal) {
	if (map.containsKey(infoKey) == false) return;

	TreeMap reflist = (TreeMap) map.get(infoKey);
	reflist.remove(infoVal.getReference());
	if (reflist.isEmpty()) map.remove(infoKey);

	update();
    }

    public synchronized String getValue(String id, String key, String ref) {
	return getValue(new EntryKey(id, key), ref);
    }

    public synchronized String getValue(EntryKey key, String ref) {
	if (map.containsKey(key) == false) return null;
	TreeMap reflist = (TreeMap) map.get(key);
	if (reflist.containsKey(ref) == false) return null;
	return ((EntryVal) reflist.get(ref)).getValue();
	
    }

    public synchronized Date getTimestamp(String id, String key, String ref) {
	Date ret = getTimestamp(new EntryKey(id, key), ref);
	return ret;
    }

    public synchronized Date getTimestamp(EntryKey key, String ref) {
	if (map.containsKey(key) == false) return null;
	TreeMap reflist = (TreeMap) map.get(key);
	if (reflist.containsKey(ref) == false) return null;
	return ((EntryVal) reflist.get(ref)).getTimestamp();
    }

    public synchronized void clear() {
	map = new TreeMap();
    }

    // implmentation du clonage "en profondeur" pour viter le clonage "superficiel".
    //  noter pour l'implmentation : les instances de Date, EntryKey, EntryVal et String ne
    // changent pas de contenu en cours d'excution (i.e. immutables).
    public synchronized Object clone() {
	State state = null;

	try {
	    state = (State) super.clone();
	}
	catch (CloneNotSupportedException ex) {
	    Debug.logExceptionAndExit(this, ex);
	}

	state.map = (TreeMap) map.clone();

	for (Iterator it = state.map.keySet().iterator(); it.hasNext();) {
	    EntryKey map_key = (EntryKey) it.next();
	    state.map.put(map_key, ((TreeMap) state.map.get(map_key)).clone());
	}

	state.date = date;

	return state;
    }

    public synchronized String dump() {
	String dump_str = new String("dumping State instance :\n");

	EntryKey map_key;
	for (Iterator it = map.keySet().iterator(); it.hasNext();) {
	    map_key = (EntryKey) it.next();
	    dump_str += "  - ([id=" + map_key.getIdentity() + "], [key=" + map_key.getKey() + "])\n";

	    String ref;
	    TreeMap reflist = (TreeMap) map.get(map_key);
	    synchronized (reflist) {
		for (Iterator it2 = reflist.keySet().iterator(); it2.hasNext();) {
		    ref = (String) it2.next();
		    dump_str += "    - [ref=" + ref + "]\n";

		    EntryVal entryval = (EntryVal) reflist.get(ref);
		    dump_str += "      - ([val=" +
			entryval.getValue() + "], [ref=" +
			entryval.getReference() + "], [timestamp=" +
			entryval.getTimestamp().toString() + " / " +
			new Long(entryval.getTimestamp().getTime() % 1000).toString() + " ms])\n";
		}
	    }
	}
	return dump_str;
    }
}

////////////////////////////////////////////////////////////

class Bdd extends Trace implements Serializable {
    // map is used as a lock to itself, not to its reflist values
    public final HashMap map = new HashMap(); // access by main, AcceptClients & Client threads

    // diffList is used as a lock to embedded data
    public ArrayList diffList = null;

    public final Object dbTransactionLock = new Object();

    public void registerDiffList(ArrayList dl) {
	diffList = dl;
    }

    public Date NextDate(Date date) {
	Date nextdate = new Date(Long.MAX_VALUE);

	synchronized (map) {
	    for (Iterator it = map.keySet().iterator(); it.hasNext();) {
		EntryKey map_key = (EntryKey) it.next();

		String ref;
		HashMap reflist = (HashMap) map.get(map_key);
		synchronized (reflist) {
		    for (Iterator it2 = reflist.keySet().iterator();
			 it2.hasNext();) {
			ref = (String) it2.next();

			TreeMap vallist = (TreeMap) reflist.get(ref);
			synchronized (vallist) {
			    SortedMap submap = (SortedMap) vallist.tailMap(date);

			    if (submap.size() > 0 &&
				((Date) submap.firstKey()).compareTo(date) == 0) {
				// Supprimer le premier element de submap
				if (submap.size() > 1)
				    submap = (SortedMap) vallist.tailMap(submap.keySet().toArray()[1]);
				else submap = null;
			    }

			    if (submap != null && submap.size() > 0 &&
				nextdate.compareTo((Date) submap.firstKey()) > 0)
				nextdate = (Date) submap.firstKey();
			}
		    }
		}
	    }
	}

	return nextdate;
    }

    public Date PrevDate(Date date) {
	Date prevdate = new Date(Long.MIN_VALUE);

	synchronized (map) {
	    for (Iterator it = map.keySet().iterator(); it.hasNext();) {
		EntryKey map_key = (EntryKey) it.next();

		String ref;
		HashMap reflist = (HashMap) map.get(map_key);
		synchronized (reflist) {
		    for (Iterator it2 = reflist.keySet().iterator();
			 it2.hasNext();) {
			ref = (String) it2.next();

			TreeMap vallist = (TreeMap) reflist.get(ref);
			synchronized (vallist) {
			    SortedMap submap = (SortedMap) vallist.headMap(date);

			    if (submap.size() > 0 &&
				prevdate.compareTo((Date) submap.lastKey()) < 0)
				prevdate = (Date) submap.lastKey();
			}
		    }
		}
	    }
	}

	return prevdate;
    }

    public String dump() {
	String dump_str = new String("dumping Bdd map :\n");

	EntryKey map_key;
	synchronized (map) {
	    for (Iterator it = map.keySet().iterator(); it.hasNext();) {
		map_key = (EntryKey) it.next();
		dump_str += "  - ([id=" + map_key.getIdentity() + "], [key=" + map_key.getKey() + "])\n";

		String ref;
		HashMap reflist = (HashMap) map.get(map_key);
		synchronized (reflist) {
		    for (Iterator it2 = reflist.keySet().iterator();
			 it2.hasNext();) {
			ref = (String) it2.next();
			dump_str += "    - [ref=" + ref + "]\n";

			Date timestamp;
			TreeMap vallist = (TreeMap) reflist.get(ref);
			synchronized (vallist) {
			    for (Iterator it3 = vallist.keySet().iterator();
				 it3.hasNext();) {
				timestamp = (Date) it3.next();
				EntryVal entryval = (EntryVal) vallist.get(timestamp);
				dump_str += "      - ([val=" +
				    entryval.getValue() + "], [ref=" +
				    entryval.getReference() + "], [timestamp=" +
				    entryval.getTimestamp().toString() + " / " +
				    new Long(entryval.getTimestamp().getTime() % 1000).toString() + " ms])\n";
			    }
			}
		    }
		}
	    }
	}
	return dump_str;
    }

    public void put(EntryKey infoKey, EntryVal infoVal) {
	HashMap reflist = null;
	TreeMap list = null;

	// update the BDD

	synchronized (map) {
	    if (map.containsKey(infoKey) == false) {
		Debug.debugln(this, "creating new HashMap (list of references)");
		reflist = new HashMap();
		map.put(infoKey, reflist);
	    } else {
		Debug.debugln(this, "getting HashMap (list of references)");
		reflist = (HashMap) map.get(infoKey);
	    }
	}

	synchronized (reflist) {
	    if (reflist.containsKey(infoVal.getReference()) == false) {
		Debug.debugln(this, "creating new TreeMap (list of references)");
		list = new TreeMap();
		reflist.put(infoVal.getReference(), list);
	    } else {
		Debug.debugln(this, "getting TreeMap (list of references)");
		list = (TreeMap) reflist.get(infoVal.getReference());
	    }
	}

	Debug.debugln(this,
		      "bdd[" + infoKey.getIdentity() + ";" + infoKey.getKey() + "]=(...;[" +
		      infoVal.getValue() + ";" + infoVal.getReference() + ";" +
		      DateFormat.getDateTimeInstance().format(infoVal.getTimestamp()) + "];...)");

	synchronized (list) {
	    if (list.containsKey(infoVal.getTimestamp()) == true) {
		Debug.debugln(this, "key with the same timestamp already registered");
		return; // nothing to update in the RMI clients states
	    }
	    else list.put(infoVal.getTimestamp(), infoVal);
	}

	// update the RMI clients states

	if (diffList != null) synchronized (diffList) {
	    for (int i = 0; i < diffList.size(); i++) {
		State state = (State) diffList.get(i);
		if (state != null) synchronized (state) {
		    if (state.map.containsKey(infoKey) &&
			((TreeMap) state.map.get(infoKey)).containsKey(infoVal.getReference())) {
			// a key/ref already exists for this State object

			EntryVal val = (EntryVal)
			    ((TreeMap) state.map.get(infoKey)).get(infoVal.getReference());
			if (infoVal.getTimestamp().compareTo(val.getTimestamp()) >= 0 &&
			    infoVal.getTimestamp().compareTo(state.date) <= 0)
			    state.put(infoKey, infoVal);
		    }

		    else {
			// no current key/ref for this State object
			if (infoVal.getTimestamp().compareTo(state.date) <= 0) {
			// we must extract the current state value from the database

			// warning : we have locked state and we now lock the database.
			// Thus, to avoid deadlocks, we must never lock a state instance
			// while a database object is synchronized.

			HashMap bddreflist = null;
			synchronized (map) {
			    if (map.containsKey(infoKey)) bddreflist = (HashMap) map.get(infoKey);
			}

			TreeMap bddvallist = null;
			if (bddreflist != null) synchronized (bddreflist) {
			    if (bddreflist.containsKey(infoVal.getReference()))
				bddvallist = (TreeMap) bddreflist.get(infoVal.getReference());
			}

			Date bddtimestamp = null;
			if (bddvallist != null) synchronized (bddvallist) {
			    SortedMap submap = (SortedMap) bddvallist.headMap(state.date);
			    if (submap.size() > 0) bddtimestamp = (Date) submap.lastKey();
			}

			if (bddtimestamp == null ||
			    infoVal.getTimestamp().compareTo(bddtimestamp) >= 0)
			    state.put(infoKey, infoVal);
			}
		    }
		}
	    }
	}
    }
}

////////////////////////////////////////////////////////////

class LogFiles extends Trace {
    private BufferedWriter writer = null; // access by main & Client threads
    private Transformer logTransformer = null; // access by Client thread

    // access by main thread
    public LogFiles() {
	try {
	    logTransformer = 
		TransformerFactory.newInstance().
		newTransformer(new StreamSource(new File(Conf.getNetTemplatesDir() + File.separator +
							 "bddfilt-log.xsl")));
	}
	catch (TransformerConfigurationException ex) {
	    Debug.logExceptionAndExit(this, ex);
	}
    }

    // access by main thread
    public void openLogFile(String name) {
	try {
	    writer = new BufferedWriter(new FileWriter(new File(Conf.getLogDir() + File.separator +
								name), true));
	}
	catch (IOException ex) {
	    Debug.logExceptionAndExit(this, ex);
	}
    }

    // access by main thread
    // assert : MUST BE CALLED BEFORE openLogFile() since some OSes doesn't support twice openning the same file
    public String loadLogFile(String name) {
	Debug.debugln(this, "Trying to load a persistant database");

	if (writer != null) {
	    Debug.errorln(this, "MUST BE CALLED BEFORE openLogFile() since some OSes do not support twice openning the same file");
	    System.exit(1);
	}

	StringBuffer strbuf = new StringBuffer();
	strbuf.append("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n<gnetwatch version=\"1.0\">\n");
	try {
	    BufferedReader reader = new BufferedReader(new FileReader(new File(Conf.getLogDir() +
									       File.separator + name)));
	    int c;
	    while ((c = reader.read()) != -1) strbuf.append((char) c);
	}
	catch (FileNotFoundException ex) {
	    Debug.debugln("no log file");
	    return null;
	}
	catch (IOException ex) {
	    Debug.logExceptionAndExit(this, ex);
	}
	strbuf.append("</gnetwatch>\n");

	Debug.debugln(this, "database correctly loaded (now need to be parsed)");
	return strbuf.toString();
    }

    // access by Client thread
    public void log(String logstr) {
	StringWriter str_writer = new StringWriter();

	try {
	    synchronized (logTransformer) {
		logTransformer.transform(new StreamSource(new StringReader(logstr)),
					 new StreamResult(new BufferedWriter(str_writer)));
	    }
	}
	catch (TransformerException ex) {
	    Debug.logExceptionAndExit(this, ex);
	}

	try {
	    synchronized (writer) {
		writer.write(str_writer.toString());
		writer.flush();
	    }
	}
	catch (IOException ex) {
	    Debug.logException(this, ex);
	}
    }
}

////////////////////////////////////////////////////////////

class Client extends Trace implements Runnable {
    private Bdd bdd; // access by main, AcceptClients & Client threads
    private Socket socket = null; // access by AcceptClients & Client threads
    private String content = null; // access by AcceptClients & Client threads
    private Transformer transformer; // access by AcceptClients & Client threads
    private LogFiles logFiles; // access by AcceptClients & Client threads

    // access by Client thread
    private void loop() {
	Debug.debugln(this, "loop()");

	if (socket != null) {
	    StringBuffer strbuf = new StringBuffer();
	    try {
		Reader reader =
		    new BufferedReader(new InputStreamReader(socket.getInputStream()));
		int c;
		while ((c = reader.read()) != -1) strbuf.append((char) c);
	    }
	    catch (IOException ex) {
		Debug.logException(this, ex);
		return;
	    }
	    content = strbuf.toString();
	}

	// Debug.debugln(this, "\n" + content);

	DOMResult flatResult = new DOMResult();
	try {
	    synchronized (transformer) {
		transformer.transform(new StreamSource(new StringReader(content)), flatResult);
	    }
	}
	catch (TransformerException ex) {
	    Debug.logException(this, ex);
	    return;
	}

	if (socket != null) logFiles.log(content);

	if (flatResult != null && flatResult.getNode() != null &&
	    flatResult.getNode().getFirstChild() != null) {

	    // We must ensure atomicity between the whole BDD update
	    // and any RMI access to informations extracted from the
	    // BDD (but we do not need/want to lock the whole BDD)

	    synchronized (bdd.dbTransactionLock) {

		for (Node node = flatResult.getNode().getFirstChild().getFirstChild();
		     node != null; node = node.getNextSibling()) {
		    if (!node.getNodeName().equals("assign"))
			Debug.debugln("unknown element \"node.getNodeName()\"");
		    else {
			try {
			    String attr_identity =
				node.getAttributes().getNamedItem("identity") != null ?
				node.getAttributes().getNamedItem("identity").getNodeValue() : "";
			    String attr_key =
				node.getAttributes().getNamedItem("key") != null ?
				node.getAttributes().getNamedItem("key").getNodeValue() : "";
			    String attr_value =
				node.getAttributes().getNamedItem("value") != null ?
				node.getAttributes().getNamedItem("value").getNodeValue() : "";
			    String attr_ref =
				node.getAttributes().getNamedItem("ref") != null ?
				node.getAttributes().getNamedItem("ref").getNodeValue() : "";
			    String attr_date = node.getAttributes().getNamedItem("date") != null ?
				node.getAttributes().getNamedItem("date").getNodeValue() : null;
			    String attr_time = node.getAttributes().getNamedItem("time") != null ?
				node.getAttributes().getNamedItem("time").getNodeValue() : null;
			    String attr_microsec = node.getAttributes().getNamedItem("microsec") != null ?
				node.getAttributes().getNamedItem("microsec").getNodeValue() : null;

			    long d;
			    d = (attr_date == null) ? 0 :
				DateFormat.getDateInstance(DateFormat.SHORT).
				parse(attr_date).getTime();
			    d += (attr_time == null) ? 0 :
				new SimpleDateFormat("H:m:s").parse(attr_time).getTime();
			    d += (attr_microsec == null) ? 0 : Long.parseLong(attr_microsec) / 1000;
			    Date timestamp = new Date(d);

			    bdd.put(new EntryKey(attr_identity, attr_key),
				    new EntryVal(attr_value, attr_ref, timestamp));

			} catch (ParseException ex) {
			    Debug.logException(this, ex);
			}
		    }
		}

	    }
	}
    }

    // access by AcceptClients thread
    public Client(Socket socket, Transformer transformer, Bdd bdd, LogFiles logFiles) {
	this.socket      = socket;
	this.transformer = transformer;
	this.bdd         = bdd;
	this.logFiles    = logFiles;
    }

    // access by AcceptClients thread
    public Client(String content, Transformer transformer, Bdd bdd, LogFiles logFiles) {
	this.content     = content;
	this.transformer = transformer;
	this.bdd         = bdd;
	this.logFiles    = logFiles;
    }

    // access by Client thread
    public void run() {
	Debug.debugln(this, "thread running");
	loop();
	Debug.debugln(this, "thread exiting");
    }
}

////////////////////////////////////////////////////////////

class AcceptClients extends Trace implements Runnable {
    private Bdd bdd; // access by main, AcceptClients & Client threads
    private LogFiles logFiles; // access by main, AcceptClients & Client threads
    private String initialLog; // access by main, AcceptClients & Client threads

    private void loop() {
	Debug.debugln(this, "loop()");

	Transformer flatTransformer = null;

	try {
	    flatTransformer = 
		TransformerFactory.newInstance().
		newTransformer(new StreamSource(new File(Conf.getNetTemplatesDir() + File.separator +
							 "bddfilt-flat.xsl")));
	}
	catch (TransformerConfigurationException ex) {
	    Debug.logExceptionAndExit(this, ex);
	}

	if (initialLog != null) {
	    Debug.debugln(this, "starting initial client for log file");
	    Thread thr = new Thread(new Client(initialLog, flatTransformer, bdd, logFiles), "Initial Client");
	    thr.start();
	    try {
		// Do not accept any client data before the log file is parsed
		thr.join();
	    }
	    catch (InterruptedException ex) {
		Debug.logException(this, ex);
	    }

	    Debug.debugln(bdd.dump());
	}

	ServerSocket serverSocket = null;
	try {
	    // throws IOException
	    serverSocket = new ServerSocket(Conf.getClientsPort(),
					    Conf.getClientsBacklog());
	}
	catch (IOException ex) {
	    Debug.logException(this, ex);
	    Debug.debugln(this, "Can't collect network informations");
	    return;
	}

	while (true) {
	    try {
		Socket acceptSocket = serverSocket.accept(); // throws IOException
		Debug.debugln(this, "new client connection");
		new Thread(new Client(acceptSocket, flatTransformer, bdd, logFiles), "Client").start();
	    }
	    catch (IOException ex) {
		Debug.logException(this, ex);
	    }
	}
    }

    // access by main thread
    public Thread startThread(Bdd bdd, LogFiles logFiles, String initialLog) {
	this.bdd = bdd;
	this.logFiles = logFiles;
	this.initialLog = initialLog;
	Thread thr = new Thread(this, "AcceptClients");
	thr.start();
	return thr;
    }

    // access by main thread
    public void run() {
	Debug.debugln(this, "thread running");
	loop();
	Debug.debugln(this, "thread exiting");
    }
}

////////////////////////////////////////////////////////////

class RemoteAccessImpl extends UnicastRemoteObject implements RemoteAccess {
    private final Bdd bdd;

    private static final int MAXCLIENTS = 32768;
    private final ArrayList diffList = new ArrayList();

    public RemoteAccessImpl(Bdd bdd) throws RemoteException {
	super();
	this.bdd = bdd;
	bdd.registerDiffList(diffList);
    }

    private State getState(Integer id) {
 	try {
	    synchronized (diffList) {
		return (State) diffList.get(id.intValue());
	    }
	}
	catch (IndexOutOfBoundsException ex) {
	    Debug.errorln("RMI state: invalid id : " + id.toString());
	}
	return null;
    }

    public Integer Connect() throws RemoteException {
	int i;

	synchronized (diffList) {
	    for (i = 0; i < diffList.size(); i++) {
		if (diffList.get(i) == null) {
		    diffList.add(i, new State());

		    Debug.debugln("RMI state: Connect from client #" + new Integer(i).toString());
		    return new Integer(i);
		}
	    }

	    if (diffList.size() == MAXCLIENTS) {
		Debug.errorln("RMI Connect: too many clients");
		return new Integer(-1);
	    }

	    diffList.add(new State());
	    Debug.debugln("RMI state: Connect from client #" + new Integer(diffList.size() - 1).toString());
	    return new Integer(diffList.size() - 1);
	}
    }

    public void Disconnect(Integer id) throws RemoteException {
	Debug.debugln("RMI state: Disconnect from client #" + id.toString());

	synchronized (diffList) {
	    try {
		diffList.set(id.intValue(), null);
	    }
	    catch (IndexOutOfBoundsException ex) {
		Debug.errorln("RMI Disconnect: invalid id : " + id.toString());
	    }
	}
    }

    public Date PrevDate(Integer id) throws RemoteException {
	State state = getState(id);
	if (state == null) return null;

	Date timestamp;
	synchronized (state) {
	    timestamp = state.date;
	}

	Date prevdate = bdd.PrevDate(timestamp);
	return SetDate(id, prevdate);
    }

    public Date NextDate(Integer id) throws RemoteException {
	State state = getState(id);
	if (state == null) return null;

	Date timestamp;
	synchronized (state) {
	    timestamp = state.date;
	}

	Date nextdate = bdd.NextDate(timestamp);
	return SetDate(id, nextdate);
    }

    public Date SetDate(Integer id, Date new_timestamp) throws RemoteException {
	State state = getState(id);
	if (state == null) return null;

	Debug.debugln("-------------------- BDD.DUMP(): " + bdd.dump());

	synchronized (bdd.dbTransactionLock) {
	    synchronized (state) {
		// warning : we have locked state and we will further lock the database.
		// Thus, to avoid deadlocks, we must never lock a state instance
		// while a database object is synchronized.

		synchronized(bdd.map) {
		    EntryKey map_key;
		    for (Iterator it = bdd.map.keySet().iterator(); it.hasNext();) {
			map_key = (EntryKey) it.next();

			String ref;
			HashMap reflist = (HashMap) bdd.map.get(map_key);
			synchronized (reflist) {
			    for (Iterator it2 = reflist.keySet().iterator(); it2.hasNext();) {
				ref = (String) it2.next();

				TreeMap vallist = (TreeMap) reflist.get(ref);
				synchronized (vallist) {
				    EntryVal bdd_new_val;
				    boolean cancel_val = false;
				    boolean same_val = true; // same_val is valid when bdd_new_val is not null

				    if (vallist.containsKey(new_timestamp))
					bdd_new_val = (EntryVal) vallist.get(new_timestamp);
				    else {
					SortedMap headlist = vallist.headMap(new_timestamp);
					if (headlist.size() > 0)
					    bdd_new_val = (EntryVal) headlist.get(headlist.lastKey());
					else {
					    bdd_new_val = null;

					    // we need to check if there is a current value in the database
					    // because this means that the value must be cleared from the current state
					    // even if it is not in the diff

					    if (vallist.get(state.date) != null || vallist.headMap(state.date).size() > 0)
						cancel_val = true;
					}
				    }

				    if (bdd_new_val != null) {
					// check if bdd_new_val and the current bdd value for the state timestamp are the same


					if (vallist.get(state.date) != null) {
					    if (!((EntryVal) vallist.get(state.date)).getValue().equals(bdd_new_val.getValue()))
						same_val = false;
					    else same_val = true;
					} else if ((vallist.headMap(state.date).size() > 0)) {
					    if (!((EntryVal) vallist.get(vallist.headMap(state.date).lastKey())).getValue().equals(bdd_new_val.getValue()))
						same_val = false;
					    else same_val = true;
					} else same_val = false;
				    }

				    if (state.map.containsKey(map_key)) {
					// l'objet etat contient deja cette cle

					TreeMap state_reflist = (TreeMap) state.map.get(map_key);
					if (state_reflist.containsKey(ref)) {
					    // la ref est associee a cette cle dans l'objet etat

					    if (bdd_new_val != null && same_val == false) state.put(map_key, bdd_new_val);
					    else state.put(map_key, new EntryVal("", ref, new_timestamp));
					} else {
					    // aucune ref n'est associee a cette cle dans l'objet etat

					    if (bdd_new_val != null && same_val == false) state.put(map_key, bdd_new_val);
					    else {
						if (cancel_val == true)
						    state.put(map_key, new EntryVal("", ref, new_timestamp));
					    }
					}

				    } else // l'objet etat ne contient pas cette cle

					if (bdd_new_val != null && same_val == false) state.put(map_key, bdd_new_val);
					else {
					    if (cancel_val == true)
						state.put(map_key, new EntryVal("", ref, new_timestamp));
					}
				}
			    }
			}
		    }
		}

		state.date = new_timestamp;
		state.update();
	    }
	}

	return new_timestamp;
    }

    public Date GetDate(Integer id) throws RemoteException {
	State state = getState(id);
	if (state == null) return null;

	Date ret_date;
	synchronized (state) {
	    ret_date = state.date;
	}
	return ret_date;
    }

    public State GetDiff(Integer id) throws RemoteException, InterruptedException {
	// we need to make a transaction because we want
	// atomicity between the modifications into the BDD coming
	// from a single client and the diff we fetch
	State state = getState(id);
	if (state == null) return null;

	// wait for an update
	boolean is_updated = state.waitForUpdate();
	if (is_updated == false) return null;
	else synchronized (bdd.dbTransactionLock) {

	    // On est oblige de cloner en profondeur car on ne peut
	    // pas garder le verrou sur state jusqu'a ce que le
	    // message soit renvoye au client RMI (c'est intrinseque
	    // au modele evenementiel de RMI) D'autre part, si le
	    // client et le serveur sont dans le meme processus
	    // (i.e. sans RMI), il faut obligatoirement dupliquer
	    // l'objet pour avoir le meme comportement qu'avec RMI.

	    // Aucune modification a l'etat ne peut intervenir entre
	    // l'execution des deux lignes qui suivent car les
	    // modifications d'une instance State n'interviennent que
	    // lorsque dbTransactionLock est verrouill. Le
	    // synchronized(state) qui suit est donc inutile.

	    State ret;
	    synchronized (state) {
		ret = (State) state.clone();
		state.clear();
	    }

	    return ret;
	}
    }

    public ArrayList GetHistory(String identity, String key, String reference) throws RemoteException {
	synchronized (bdd.map) {
	    HashMap reflist =  (HashMap) bdd.map.get(new EntryKey(identity, key));
	    if (reflist == null) return null;

	    synchronized (reflist) {
		TreeMap vallist = (TreeMap) reflist.get(reference);
		if (vallist == null) return null;
		return new ArrayList(vallist.values());
	    }
	}
    }

    public String EchoString(String str) throws RemoteException {
	return "[" + str + "]";
    }

    public String DumpDatabase() throws RemoteException {
	return bdd.dump();
    }

    public void Ping() throws RemoteException {
	return;
    }

    public Bdd Tst() throws RemoteException {
	Debug.debugln("RemoteExecution : Tst()");
	return bdd;
    }
}

////////////////////////////////////////////////////////////

class IntroPanel extends JPanel {
    private final Image img;

    public IntroPanel(String img_name) {
	setBackground(Color.black);

	MediaTracker tracker = new MediaTracker(this);
	img = Toolkit.getDefaultToolkit().getImage(Conf.getNetTemplatesDir() +
						   File.separator + img_name);
	tracker.addImage(img, 0);
	try {
	    tracker.waitForID(0);
	}
	catch (InterruptedException ex) {
	    Debug.logExceptionAndExit(this, ex);
	}
    }

    public void paintComponent(Graphics g) {
	super.paintComponent(g);
	Graphics2D g2d = (Graphics2D) g;
	g2d.drawImage(img, 0, 0, this);
    }

    public Dimension getDimension() {
	return new Dimension(img.getWidth(this), img.getHeight(this));
    }
}

class View extends JPanel {
    private final Image img;

    public View(String img_name) {
	setBackground(Color.black);

	MediaTracker tracker = new MediaTracker(this);
	img = Toolkit.getDefaultToolkit().getImage(Conf.getNetTemplatesDir() + File.separator + img_name);
	tracker.addImage(img, 0);
	try {
	    tracker.waitForID(0);
	}
	catch (InterruptedException ex) {
	    Debug.logExceptionAndExit(this, ex);
	}
    }

    public void paintComponent(Graphics g) {
	super.paintComponent(g);
	Graphics2D g2d = (Graphics2D) g;
	g2d.drawImage(img, 0, 0, this);
    }

    public void informNewShape(JComponent component, Point old_loc, Dimension old_dim) {
	Point new_loc = component.getLocation();
	Dimension new_dim = new Dimension(component.getWidth(), component.getHeight());

	int old_max_w = old_loc.x + old_dim.width;
	int old_max_h = old_loc.y + old_dim.height;
	int new_max_w = new_loc.x + new_dim.width;
	int new_max_h = new_loc.y + new_dim.height;

	int view_max_w = getWidth();
	int view_max_h = getHeight();

	int new_w = 0;
	int new_h = 0;

	if ((view_max_w == old_max_w && old_max_w != new_max_w) || // le composant etait potentiellement le plus a droite
	    (view_max_w < new_max_w) || // le composant est maintenant le plus a droite
	    (view_max_h == old_max_h && old_max_h != new_max_h) || // le composant etait potentiellement le plus bas
	    (view_max_h < new_max_h)) { // le composant est maintenant le plus bas
	    for (int i = 0; i < getComponentCount(); i++) {
		Rectangle rect = new Rectangle();
		getComponent(i).getBounds(rect);
		if (rect.x + rect.width > new_w) new_w = rect.x + rect.width;
		if (rect.y + rect.height > new_h) new_h = rect.y + rect.height;
	    }
	    setPreferredSize(new Dimension(new_w, new_h));
	    revalidate();
	}
    }
}

////////////////////////////////////////////////////////////

interface WantNotifyReshape {
    public void notifyReshape(VGraphic graphic);
}

////////////////////////////////////////////////////////////

class VGraphic extends JComponent {
    private static final int NSTEPS = 20;

    private final View view;
    private final ArrayList wantNotifyReshapeList = new ArrayList();
    private int preferredX = 0;
    private int preferredY = 0;
    private int stepX = 0;
    private int stepY = 0;

    public VGraphic(View view) {
	this.view = view;
    }

    public void addNotifyReshape(WantNotifyReshape obj) {
	if (wantNotifyReshapeList.contains(obj) == false)
	    wantNotifyReshapeList.add(obj);
    }

    public void removeNotifyReshape(WantNotifyReshape obj) {
	wantNotifyReshapeList.remove(obj);
    }

    public void destroy() {}

    public void notifyReshape() {
	for (int i = 0; i < wantNotifyReshapeList.size(); i++)
	    ((WantNotifyReshape) wantNotifyReshapeList.get(i)).notifyReshape(this);
    }

    public void setSizeAndInform(int width, int height) {
	if (width == getWidth() && height == getHeight()) return;

	Dimension before_dim = new Dimension(getWidth(), getHeight());
	setSize(width, height);
	view.informNewShape(this, getLocation(), before_dim);
	notifyReshape();
    }

    public void setPreferredLocation(int x, int y) {
	if (isVisible() == false) {
	    setVisible(true);
	    preferredX = x;
	    preferredY = y;
	    stepX = 0;
	    stepY = 0;
	    setLocationAndInform(x, y);
	    return;
	}

	preferredX = x;
	preferredY = y;

	stepX = (preferredX - getLocation().x) / NSTEPS;
	if (stepX == 0)
	    stepX = preferredX > getLocation().x ? 1 : -1;

	stepY = (preferredY - getLocation().y) / NSTEPS;
	if (stepY == 0)
	    stepY = preferredY > getLocation().y ? 1 : -1;
    }

    public void moveToPreferredLocationAndInform() {
	if (stepX == 0 && stepY == 0) return;

	int new_x = getLocation().x + stepX;
	if ((getLocation().x <= preferredX && new_x > preferredX) ||
	    (getLocation().x >= preferredX && new_x < preferredX))
	    new_x = preferredX;

	int new_y = getLocation().y + stepY;
	if ((getLocation().y <= preferredY && new_y > preferredY) ||
	    (getLocation().y >= preferredY && new_y < preferredY))
	    new_y = preferredY;

	if (new_x == preferredX) stepX = 0;
	if (new_y == preferredY) stepY = 0;

	setLocationAndInform(new_x, new_y);
    }

    public void setLocationAndInform(int x, int y) {
	if (x == getLocation().x && y == getLocation().y) return;

	Point before_pt = getLocation();
	setLocation(x, y);
	view.informNewShape(this, before_pt, new Dimension(getWidth(), getHeight()));
	notifyReshape();
    }

    public void setBoundsAndInform(int x, int y, int width, int height) {
	if (x == getLocation().x && y == getLocation().y &&
	    width == getWidth() && height == getHeight()) return;

	Point before_pt = getLocation();
	Dimension before_dim = new Dimension(getWidth(), getHeight());
	reshape(x, y, width, height);
	view.informNewShape(this, before_pt, before_dim);
	notifyReshape();
    }
}

////////////////////////////////////////////////////////////

class VText extends VGraphic implements WantNotifyReshape {
    private static final Font font = new Font("Helvetica", Font.BOLD, 12);
    private String text;
    private int text_width = 0;
    private VGraphic graphic;

    public VText(View view, VGraphic graphic, String text) {
	super(view);
	this.text = text;
	this.graphic = graphic;
	graphic.addNotifyReshape(this);
    }

    public void destroy() {
	graphic.removeNotifyReshape(this);
    }

    public void addNotify() {
	super.addNotify();

	Graphics2D g2d = (Graphics2D) getGraphics();

	g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
	g2d.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);

	FontRenderContext frc = g2d.getFontRenderContext();
	TextLayout tl = new TextLayout(text, font, frc);
	AffineTransform transform = new AffineTransform();

	Shape shape = tl.getOutline(transform);
	text_width = (int) shape.getBounds().getWidth();
	setSizeAndInform((int) shape.getBounds().getWidth(), (int) shape.getBounds().getHeight());
    }

    public void paintComponent(Graphics g) {
	super.paintComponent(g);
	Graphics2D g2d = (Graphics2D) g;

	g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
	g2d.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);

	FontRenderContext frc = g2d.getFontRenderContext();
	TextLayout tl = new TextLayout(text, font, frc);
	AffineTransform transform = new AffineTransform();

	transform.translate(0, tl.getBounds().getHeight());
	Shape shape = tl.getOutline(transform);

 	g2d.setColor(Color.RED);
 	g2d.fill3DRect(0, 0, 1000, 1000, false);
	g2d.setColor(Color.BLACK);
	g2d.fill(shape);
    }

    public void notifyReshape(VGraphic graphic) {
 	Rectangle rect = new Rectangle();
 	graphic.getBounds(rect);
 	setLocationAndInform(rect.x + (rect.width - text_width) / 2, rect.y + rect.height + 2);
    }
}

////////////////////////////////////////////////////////////

class VConnector extends VGraphic implements WantNotifyReshape {
    private static final int border_width = 20;
    private VGraphic graphic1, graphic2;
    private boolean first_diag = true;

    private static final String PARAM_PLAIN = "plain";

    private String param_type = PARAM_PLAIN;
    private int param_thick = 5;
    private Color param_col1 = Color.black;
    private Color param_col2 = Color.blue;

    public VConnector(View view, VGraphic graphic1, VGraphic graphic2) {
	super(view);
	this.graphic1 = graphic1;
	this.graphic2 = graphic2;
	graphic1.addNotifyReshape(this);
	graphic2.addNotifyReshape(this);
    }

    public VConnector(View view, VGraphic graphic1, VGraphic graphic2, Color col1, Color col2, int thick) {
	super(view);
	this.graphic1 = graphic1;
	this.graphic2 = graphic2;
	param_col1 = col1;
	param_col2 = col2;
	param_thick = thick;
	graphic1.addNotifyReshape(this);
	graphic2.addNotifyReshape(this);
    }

    public void destroy() {
	graphic1.removeNotifyReshape(this);
	graphic2.removeNotifyReshape(this);
    }

    public void updateThick(int thick) {
	param_thick = thick;
	repaint();
    }

    public void updateColors(Color col1, Color col2) {
	param_col1 = col1;
	param_col2 = col2;
	repaint();
    }
    public void paintComponent(Graphics g) {
	super.paintComponent(g);
	Graphics2D g2d = (Graphics2D) g;

	g2d.setStroke(new BasicStroke(param_thick, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));

	int x1, y1, x2, y2;

	if (first_diag == true) {
	    x1 = border_width;
	    y1 = border_width;
	    x2 = getWidth() - 1 - border_width;
	    y2 = getHeight() - 1 - border_width;
	} else {
	    x1 = getWidth() - 1 - border_width;
	    y1 = border_width;
	    x2 = border_width;
	    y2 =  getHeight() - 1 - border_width;
	}

	g2d.setPaint(new GradientPaint(new Point2D.Double(x1, y1),
				       param_col1,
				       new Point2D.Double(x2, y2),
				       param_col2));
	//	g2d.setColor(Color.white);

	g2d.drawLine(x1, y1, x2, y2);

// 	Debug.debugln("paint vconnector " + new Integer(getLocation().x).toString() + " " +
// 		      new Integer(getLocation().y).toString() + " " +
// 		      new Integer(getWidth()).toString() + " " +
// 		      new Integer(getHeight()).toString());
    }

    public void notifyReshape(VGraphic graphic) {
 	Rectangle rect1 = new Rectangle(), rect2 = new Rectangle();
 	graphic1.getBounds(rect1);
 	graphic2.getBounds(rect2);

 	Point center1 = new Point(rect1.x + rect1.width / 2, rect1.y + rect1.height / 2);
 	Point center2 = new Point(rect2.x + rect2.width / 2, rect2.y + rect2.height / 2);

 	Point top_left = new Point();
 	Point bottom_right = new Point();

	if (center1.x < center2.x) {
	    top_left.x = center1.x;
	    bottom_right.x = center2.x;
	} else {
	    top_left.x = center2.x;
	    bottom_right.x = center1.x;
	}

	if (center1.y < center2.y) {
	    top_left.y = center1.y;
	    bottom_right.y = center2.y;
	} else {
	    top_left.y = center2.y;
	    bottom_right.y = center1.y;
	}

	if (center1.x < center2.x && center1.y < center2.y ||
	    center1.x > center2.x && center1.y > center2.y)
	    first_diag = true;
	else first_diag = false;

	setBoundsAndInform(top_left.x - border_width, top_left.y - border_width,
			   bottom_right.x - top_left.x + 1 + 2 * border_width,
			   bottom_right.y - top_left.y + 1 + 2 * border_width);
    }
}

////////////////////////////////////////////////////////////

class VPict extends VGraphic implements MouseListener, MouseMotionListener {
    private Image img;
    private boolean mouse_dragged = false;

    public VPict(View view, String img_name) {
	super(view);

	addMouseListener(this);
	addMouseMotionListener(this);

	MediaTracker tracker = new MediaTracker(this);
	img = Toolkit.getDefaultToolkit().getImage(Conf.getNetTemplatesDir() + File.separator + img_name);
	tracker.addImage(img, 0);
	try {
	    tracker.waitForID(0);
	}
	catch (InterruptedException ex) {
	    Debug.logExceptionAndExit(this, ex);
	}
    }

    public void addNotify() {
	super.addNotify();
	setSizeAndInform(img.getWidth(this), img.getHeight(this));
    }

    public void changeShape(String img_name) {
	MediaTracker tracker = new MediaTracker(this);
	img = Toolkit.getDefaultToolkit().getImage(Conf.getNetTemplatesDir() + File.separator + img_name);
	tracker.addImage(img, 0);
	try {
	    tracker.waitForID(0);
	}
	catch (InterruptedException ex) {
	    Debug.logExceptionAndExit(this, ex);
	}

	repaint();
    }

    public void paintComponent(Graphics g) {
	super.paintComponent(g);
	Graphics2D g2d = (Graphics2D) g;
	if (mouse_dragged == true)
	    g2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, (float) .5));
	g2d.drawImage(img, 0, 0, this);
    }

    // interface MouseListener
    public void mouseClicked(MouseEvent e) {/*Debug.debugln("click");*/}
    public void mouseEntered(MouseEvent e) {/*Debug.debugln("enter");*/}
    public void mouseExited(MouseEvent e) {/*Debug.debugln("exit");*/}

    public void mousePressed(MouseEvent e) {
    	mouse_dragged = true;
	repaint();
    }

    public void mouseReleased(MouseEvent e) {
	mouse_dragged = false;
	repaint();
    }

    // interface MouseMotionListener
    public void mouseDragged(MouseEvent e) {
// 	Debug.debugln("new mouse location : " +
// 		      new Integer(e.getX()).toString() + " " + new Integer(e.getY()).toString());
	setLocationAndInform(e.getX() + getX() - getWidth() / 2, e.getY() + getY() - getHeight() / 2);
    }
    public void mouseMoved(MouseEvent e) {/*Debug.debugln("move");*/}
}

////////////////////////////////////////////////////////////

class ViewerDisplay extends Trace {
    final Viewer viewer;

    private static final int ANIM_MSEC = 20;

    private final static String XML_EMPTY = "";
    private final static String XML_CLASS = "class";
    private final static String XML_BASIC_COMPONENT = "basic component";
    private final static String XML_XPOS = "xpos";
    private final static String XML_YPOS = "ypos";
    private final static String XML_SHAPE = "shape";
    private final static String XML_DESCR = "descr";
    private final static String XML_INFO = "info";
    private final static String XML_LINK = "link";

    private javax.swing.Timer moving_timer = null;

    // access by Swing thread
    private HashMap map_vpict = new HashMap();
    private HashMap map_vtext = new HashMap();
    private HashMap map_vconnector = new HashMap();

    public ViewerDisplay(Viewer viewer) {
	this.viewer = viewer;

	Action actionMoveTimer = new AbstractAction() {
		public void actionPerformed(ActionEvent e) {
		    for (Iterator it = map_vpict.values().iterator(); it.hasNext();) {
			((VPict) it.next()).moveToPreferredLocationAndInform();
		    }
		}
	    };

	moving_timer = new javax.swing.Timer(ANIM_MSEC, actionMoveTimer);
	moving_timer.start();
    }

    // access by Viewer thread
    public void informState(final State state, final State diff) {
	final Date date;

	Debug.debugln("-------------------- STATE.DUMP() : " + state.dump());
	Debug.debugln("-------------------- DIFF.DUMP() : " + diff.dump());

	synchronized (state) { date = state.date; }

	try {
	    SwingUtilities.invokeAndWait(new Runnable() {
		    public void run() {
			if (viewer.menu_windows_startdebug.getState() == true)
			    viewer.addMessage("<P><B>Dumping state</B><BR><FONT SIZE=-2>" + state.dump() +
					      "</FONT><BR><B>Dumping diff</B><BR><FONT SIZE=-2>" + diff.dump() +
					      "</FONT>");

			viewer.displayDate(date);
			viewer.displayKeysTree(state);
			updateViewer(state, diff);
		    }
		});
	}
	catch (InterruptedException ex) {
	    Debug.logExceptionAndExit(ex);
	}
	catch (InvocationTargetException ex) {
	    Debug.logExceptionAndExit(ex);
	}
    }

    // access by Swing thread
    private void updateViewer(final State state, final State diff) {
	synchronized (diff) {
	    for (Iterator it = diff.map.keySet().iterator(); it.hasNext();) {
		EntryKey map_key = (EntryKey) it.next();
		TreeMap reflist = (TreeMap) diff.map.get(map_key);
		synchronized (reflist) {
		    for (Iterator it2 = reflist.keySet().iterator(); it2.hasNext();)
			updateIdKeyRefVal(map_key, (EntryVal) reflist.get((String) it2.next()), state);
		}
	    }
	}
    }

    // access by Swing thread - diff is locked, state is not locked
    private void updateIdKeyRefVal(EntryKey entry_key, EntryVal entry_val, final State state) {
	Debug.debugln("-------------------- update IdKeyRefVal : " + entry_key.dump() + " - " + entry_val.dump());

	if (entry_val.getReference().equals("")) {
	    // these update methods are called only if it doesn't concern a link
	    if (entry_key.getKey().equals(XML_XPOS)) updateXPOS(entry_key, entry_val, state);
	    if (entry_key.getKey().equals(XML_YPOS)) updateYPOS(entry_key, entry_val, state);
	    if (entry_key.getKey().equals(XML_SHAPE)) updateSHAPE(entry_key, entry_val, state);
	    if (entry_key.getKey().equals(XML_DESCR)) updateDESCR(entry_key, entry_val, state);
	    if (entry_key.getKey().equals(XML_INFO)) updateINFO(entry_key, entry_val, state);
	} else {
	    // these update method is called only if it concerns a link
	    if (entry_key.getKey().startsWith(XML_LINK + ",")) updateLINK(entry_key, entry_val, state);
	}
    }

    // if vpict exists : return vpict
    // if vpict doesn't exist : create vpict and vtext
    private VPict getVPict(String identity, final State state) {
	if (map_vpict.containsKey(identity) == false) {
	    String shape = state.getValue(identity, XML_SHAPE, XML_EMPTY);
	    if (shape == null || shape.equals(XML_EMPTY)) return null;

	    String xpos = state.getValue(identity, XML_XPOS, XML_EMPTY);
	    if (xpos == null || xpos.equals(XML_EMPTY)) return null;

	    String ypos = state.getValue(identity, XML_YPOS, XML_EMPTY);
	    if (ypos == null || ypos.equals(XML_EMPTY)) return null;

	    VPict vpict = new VPict(viewer.getView(), shape + ".png");
	    vpict.setVisible(false);

	    String info = state.getValue(identity, XML_INFO, XML_EMPTY);
	    if (info != null && info.equals("") == false)
		vpict.setToolTipText(info);

	    map_vpict.put(identity, vpict);
	    viewer.getView().add(vpict);

	    // get vtext
	    if (map_vtext.containsKey(identity) == false) {
		String descr = state.getValue(identity, XML_DESCR, XML_EMPTY);
		if (descr == null || descr.equals(XML_EMPTY)) return null;

		VText vtext = new VText(viewer.getView(), vpict, descr);
		map_vtext.put(identity, vtext);
		viewer.getView().add(vtext);
		vpict.notifyReshape();
	    }

	    return vpict;
	} else return (VPict) map_vpict.get(identity);
    }

    // remove vpict
    private void removeVPict(String identity, final State state) {
	if (map_vpict.containsKey(identity) == false) return;
	VPict vpict = (VPict) map_vpict.get(identity);
	vpict.destroy();
	map_vpict.remove(identity);
	viewer.getView().remove(vpict);
	viewer.getView().repaint();
    }

    // return vtext (if vpict doesn't exist : create vpict and vtext)
    private VText getVTextForVPict(String identity, final State state) {
	VPict vpict = getVPict(identity, state);
	if (vpict == null) return null;

	if (map_vtext.containsKey(identity) == false) return null;
	return (VText) map_vtext.get(identity);
    }

    // remove vtext
    private void removeVTextForVPict(String identity, final State state) {
	if (map_vtext.containsKey(identity) == false) return;
	VText vtext = (VText) map_vtext.get(identity);
	vtext.destroy();
	map_vtext.remove(identity);
	viewer.getView().remove(vtext);
	viewer.getView().repaint();
    }

    // create the two vpict related nodes if possible
    // create the vconnector if it doesn't exist
    private VConnector getVConnector(String id1, String id2, String link_key, final State state) {
	if (map_vconnector.containsKey(id1) == true)
	    return (VConnector) map_vconnector.get(id1);
	if (map_vconnector.containsKey(id2) == true)
	    return (VConnector) map_vconnector.get(id2);

	// create vconnector
	VPict vpict1 = 	getVPict(id1, state);
	VPict vpict2 = 	getVPict(id2, state);
	if (vpict1 == null || vpict2 == null) return null;

	String link_val_1 = state.getValue(id1, XML_LINK + "," + link_key, id2);
	String link_val_2 = state.getValue(id2, XML_LINK + "," + link_key, id1);

	String link_val = null;
	if (link_val_1 == null || link_val_1.equals("")) {
	    if (link_val_2 == null || link_val_2.equals("")) return null;
	    link_val = link_val_2;
	} else
	    if (link_val_2 == null || link_val_2.equals(""))
		link_val = link_val_1;
	    else {
		if (state.getTimestamp(id1,
				       XML_LINK + "," + link_key,
				       id2).compareTo(state.getTimestamp(id2,
								     XML_LINK + "," + link_key,
								     id1)) >= 0)
		    link_val = link_val_1;
		else link_val = link_val_2;
	    }	    

	String [] params = link_val.split(",");
	if (params.length < 9) return null;

	Color col1 = new Color(Integer.valueOf(params[1]).intValue(),
			       Integer.valueOf(params[2]).intValue(),
			       Integer.valueOf(params[3]).intValue());
	Color col2 = new Color(Integer.valueOf(params[4]).intValue(),
			       Integer.valueOf(params[5]).intValue(),
			       Integer.valueOf(params[6]).intValue());

	int thick = Integer.valueOf(params[7]).intValue();

	VConnector vconnector = new VConnector(viewer.getView(), vpict1, vpict2, col1, col2, thick);
	map_vconnector.put(id1, vconnector);
	map_vconnector.put(id2, vconnector);
	viewer.getView().add(vconnector);
	vpict1.notifyReshape();
	return vconnector;
    }

    private void removeVConnector(String identity, final State state) {
	if (map_vconnector.containsKey(identity) == false) return;
	VConnector vconnector = (VConnector) map_vconnector.get(identity);
	vconnector.destroy();
	map_vconnector.remove(identity);
	viewer.getView().remove(vconnector);
	viewer.getView().repaint();
    }

    private void updateXPOS(EntryKey entry_key, EntryVal entry_val, final State state) {
	if (entry_val.getValue().equals(XML_EMPTY)) {
	    removeVPict(entry_key.getIdentity(), state);
	    removeVTextForVPict(entry_key.getIdentity(), state);
	    removeVConnector(entry_key.getIdentity(), state);
	    return;
	}

	String ypos = state.getValue(entry_key.getIdentity(), XML_YPOS, XML_EMPTY);
	if (ypos == null || ypos.equals(XML_EMPTY)) {
	    removeVPict(entry_key.getIdentity(), state);
	    removeVTextForVPict(entry_key.getIdentity(), state);
	    removeVConnector(entry_key.getIdentity(), state);
	    return;
	}

	VPict vpict = getVPict(entry_key.getIdentity(), state);
	if (vpict == null) return;
 	vpict.setPreferredLocation(Integer.valueOf(entry_val.getValue()).intValue(),
 				   Integer.valueOf(ypos).intValue());
    }

    private void updateYPOS(EntryKey entry_key, EntryVal entry_val, final State state) {
	if (entry_val.getValue().equals(XML_EMPTY)) {
	    removeVPict(entry_key.getIdentity(), state);
	    removeVTextForVPict(entry_key.getIdentity(), state);
	    removeVConnector(entry_key.getIdentity(), state);
	    return;
	}

	String xpos = state.getValue(entry_key.getIdentity(), XML_XPOS, XML_EMPTY);
	if (xpos == null || xpos.equals(XML_EMPTY)) {
	    removeVPict(entry_key.getIdentity(), state);
	    removeVTextForVPict(entry_key.getIdentity(), state);
	    removeVConnector(entry_key.getIdentity(), state);
	    return;
	}

	VPict vpict = getVPict(entry_key.getIdentity(), state);
	if (vpict == null) return;
 	vpict.setPreferredLocation(Integer.valueOf(xpos).intValue(),
 				   Integer.valueOf(entry_val.getValue()).intValue());
    }

    private void updateSHAPE(EntryKey entry_key, EntryVal entry_val, final State state) {
	VPict vpict = getVPict(entry_key.getIdentity(), state);
	if (vpict == null) return;

	vpict.changeShape(entry_val.getValue() + ".png");
    }

    private void updateINFO(EntryKey entry_key, EntryVal entry_val, final State state) {
	VPict vpict = getVPict(entry_key.getIdentity(), state);
	if (vpict == null) return;

	String info = state.getValue(entry_key.getIdentity(), XML_INFO, XML_EMPTY);
	if (info != null && info.equals("") == false)
	    vpict.setToolTipText(info);
    }

    private void updateDESCR(EntryKey entry_key, EntryVal entry_val, final State state) {
	String descr = entry_val.getValue();

	if (descr.equals("")) {
	    removeVTextForVPict(entry_key.getIdentity(), state);
	    return;
	}

	VPict vpict = getVPict(entry_key.getIdentity(), state);
	if (vpict == null) return;
	VText vtext = getVTextForVPict(entry_key.getIdentity(), state);
	if (vtext != null) removeVTextForVPict(entry_key.getIdentity(), state);
	vtext = new VText(viewer.getView(), vpict, descr);
	map_vtext.put(entry_key.getIdentity(), vtext);
	viewer.getView().add(vtext);
	vpict.notifyReshape();
    }

    private void updateLINK(EntryKey entry_key, EntryVal entry_val, final State state) {
	String id1 = entry_key.getIdentity();
	String id2 = entry_val.getReference();

	String [] keys = entry_key.getKey().split(",");
	if (keys.length < 2) return;

	VConnector vconnector = getVConnector(id1, id2, keys[1], state);
	if (vconnector == null) return;

// 	// since vconnector is not null, vpict1 and vpict2 must exist
// 	VPict vpict1 = 	getVPict(id1, state);
// 	VPict vpict2 = 	getVPict(id2, state);

	String link_val_1 = state.getValue(id1, XML_LINK + "," + keys[1], id2);
	String link_val_2 = state.getValue(id2, XML_LINK + "," + keys[1], id1);

	String link_val = null;
	if (link_val_1 == null || link_val_1.equals("")) {
	    if (link_val_2 == null || link_val_2.equals("")) link_val = null;
	    else link_val = link_val_2;
	} else
	    if (link_val_2 == null || link_val_2.equals(""))
		link_val = link_val_1;
	    else {
		if (state.getTimestamp(id1,
				       XML_LINK + "," + keys[1],
				       id2).compareTo(state.getTimestamp(id2,
									 XML_LINK + "," + keys[1],
									 id1)) >= 0)
		    link_val = link_val_1;
		else link_val = link_val_2;
	    }	    

	if (link_val == null) {
	    removeVConnector(id1, state);
	    removeVConnector(id2, state);
	    return;
	}

	String [] params = link_val.split(",");
	if (params.length < 9) return;

	Color col1 = new Color(Integer.valueOf(params[1]).intValue(),
			       Integer.valueOf(params[2]).intValue(),
			       Integer.valueOf(params[3]).intValue());
	Color col2 = new Color(Integer.valueOf(params[4]).intValue(),
			       Integer.valueOf(params[5]).intValue(),
			       Integer.valueOf(params[6]).intValue());

	int thick = Integer.valueOf(params[7]).intValue();

	vconnector.updateColors(col1, col2);
	vconnector.updateThick(thick);
    }

    public void informEnd() {}
}

class ViewerState extends Trace {
    // state is modified by only one thread (Viewer), but accessed by two threads (Swing + Viewer)
    private State state = new State();

    public RemoteAccess remoteAccess = null;
    public Integer remoteId = new Integer(-1);

    public ViewerState(RemoteAccess remoteAccess) {
	this.remoteAccess = remoteAccess;

	try {
	    remoteId = remoteAccess.Connect();
	}
	catch (RemoteException ex) {
	    Debug.logExceptionAndExit(ex);
	}
    }

    public void Reconnect(RemoteAccess remoteAccess) {
	this.remoteAccess = remoteAccess;
	state = new State();

	try {
	    remoteId = remoteAccess.Connect();
	}
	catch (RemoteException ex) {
	    Debug.logExceptionAndExit(ex);
	}
    }

    public Date SetDate(Date date) {
	Date ret = null;

	if (remoteAccess == null) return null;

	try {
	    ret = remoteAccess.SetDate(remoteId, date);
	    if (ret == null) {
		Debug.errorln(this, "remoteAccess.SetDate returned null");
		System.exit(1);
	    }
	}
	catch (RemoteException ex) {
	    Debug.logExceptionAndExit(ex);
	}

	return date;
    }

    public Date NextDate() {
	Date ret = null;

	if (remoteAccess == null) return null;

	try {
	    ret = remoteAccess.NextDate(remoteId);
	    if (ret == null) {
		Debug.errorln(this, "remoteAccess.NextDate returned null");
		System.exit(1);
	    }
	}
	catch (RemoteException ex) {
	    Debug.logExceptionAndExit(ex);
	}

	return ret;
    }

    public Date PrevDate() {
	Date ret = null;

	if (remoteAccess == null) return null;

	try {
	    ret = remoteAccess.PrevDate(remoteId);
	    if (ret == null) {
		Debug.errorln(this, "remoteAccess.PrevDate returned null");
		System.exit(1);
	    }
	}
	catch (RemoteException ex) {
	    Debug.logExceptionAndExit(ex);
	}

	return ret;
    }

    // update state and return a diff from previous state
    public State UpdateState() throws InterruptedException {
	State diff = null;

	if (remoteAccess == null) return null;

	try {
	    while (diff == null) {
		diff = remoteAccess.GetDiff(remoteId);
		if (Thread.currentThread().isInterrupted() == true)
		    throw new InterruptedException("polling thread interrupted");
	    }
	}
	catch (RemoteException ex) {
	    Debug.logExceptionAndExit(ex);
	}

	synchronized (state) {
	    for (Iterator it = diff.map.keySet().iterator(); it.hasNext();) {
		EntryKey map_key = (EntryKey) it.next();
		TreeMap ref_list = (TreeMap) diff.map.get(map_key);

		for (Iterator it2 = ref_list.keySet().iterator(); it2.hasNext();) {
		    String ref = (String) it2.next();
		    if (((EntryVal) ref_list.get(ref)).getValue().compareTo("") == 0)
			state.remove(map_key, (EntryVal) ref_list.get(ref));
		    else state.put(map_key, (EntryVal) ref_list.get(ref));
		}
	    }

	    state.date = diff.date;
	}

	return diff;
    }

    public State getState() {
	return state;
    }

    protected void finalize() throws Throwable {
	super.finalize();
    }
}

////////////////////////////////////////////////////////////

class IdKeyRefNode {
    public final String label, identity, key, reference;

    public IdKeyRefNode(String label, String identity, String key, String reference) {
	this.label = label;
	this.identity = identity;
	this.key = key;
	this.reference = reference;
    }

    public String toString() {
	return label;
    }
}

////////////////////////////////////////////////////////////

class Viewer implements Runnable {
    private static final int initial_width = 800;
    private static final int initial_height = 600;

    private Thread thr = null;

    private static JFrame introFrame = null;
    private static JLabel introLabel = null;

    private static JTree keysTree = null;
    private static JTable historyTable = null;
    public JSplitPane splitPane = null;
    public boolean splitPane_removed = false;
    public JToolBar toolbar = null;
    public JPanel bottomPanel = null;
    
    private RemoteAccess remoteAccess = null;

    private View view = null;
    public JFrame frame = null;
    private ViewerState viewerState = null;
    private ViewerDisplay viewerDisplay = null;
    
    private static VPict vpict;

    private JEditorPane debugText;

    private JButton toolbar_network_backward;
    private JButton toolbar_network_backward_max;
    private JButton toolbar_network_forward;
    private JButton toolbar_network_forward_max;
    private JButton toolbar_network_play;
    private JButton toolbar_network_pause;
    private JButton toolbar_file_connect;
    private JButton toolbar_file_disconnect;

    private JMenuItem menu_network_backward;
    private JMenuItem menu_network_backward_max;
    private JMenuItem menu_network_forward;
    private JMenuItem menu_network_forward_max;
    private JMenuItem menu_network_play;
    private JMenuItem menu_network_pause;
    private JMenuItem menu_file_connect;
    private JMenuItem menu_file_disconnect;

    public JCheckBoxMenuItem menu_windows_keys;
    public JCheckBoxMenuItem menu_windows_history;
    public JCheckBoxMenuItem menu_windows_toolbar;
    public JCheckBoxMenuItem menu_windows_info;
    public JCheckBoxMenuItem menu_windows_startdebug;

    public javax.swing.Timer playing_timer = null;

    JLabel timestampLabel;

    public View getView() {
	return view;
    }

    public void run() {
	Debug.debugln(this, "thread running");

	viewerState.SetDate(new Date(Long.MAX_VALUE));

	try {
	    BddPoll();
	}
	catch (InterruptedException ex) {
	    Debug.debugln(this, "remote polling : InterruptedException");
	}

	Debug.debugln(this, "thread exiting");
    }

    public void BddPoll() throws InterruptedException {
	State diff = null;

	while (true) {
	    diff = viewerState.UpdateState();
	    if (diff != null) viewerDisplay.informState(viewerState.getState(), diff);
	}
    }

    public void BuildGUI() {
	try {
	    UIManager.setLookAndFeel("javax.swing.plaf.metal.MetalLookAndFeel");
	}
	catch (ClassNotFoundException ex) {
	    Debug.logException(this, ex);
	}
	catch (InstantiationException ex) {
	    Debug.logException(this, ex);
	}
	catch (IllegalAccessException ex) {
	    Debug.logException(this, ex);
	}
	catch (UnsupportedLookAndFeelException ex) {
	    Debug.logException(this, ex);
	}

	////////////////////////////////////////////////////////////

	//         Top
	//   (JFrame) frame
	//         ...
	//     contentPane
	// (JScrollPane) scrollPane
	//     (JPanel) view
	//  (JComponent) vpict

	////////////////////////////////////////////////////////////
	// main frame

	// build frame
	frame = new JFrame("gnetwatch");
	DebugGUI.frame = frame;

	frame.addWindowListener(new WindowAdapter() {
		public void windowClosing(WindowEvent e) {
		    stopThread();
		    System.exit(0);
		}
	    });

	////////////////////////////////////////////////////////////
	// Menu

	JMenuBar menu_bar = new JMenuBar();

	////////////////////////////////////////////////////////////
	// Menu File

	JMenu menu_file = new JMenu("File");
	menu_file.setMnemonic(KeyEvent.VK_F);

	menu_file_connect = new JMenuItem(new ActionConnect("Connect", new ImageIcon("menu_file_connect.png"),
							    "Connect to a database",
							    KeyStroke.getKeyStroke(KeyEvent.VK_C, ActionEvent.ALT_MASK),
							    KeyEvent.VK_C,
							    viewerState, this));
	menu_file_connect.setEnabled(false);
	menu_file.add(menu_file_connect);

	menu_file_disconnect = new JMenuItem(new ActionDisconnect("Disconnect", new ImageIcon("menu_file_disconnect.png"),
								  "Disconnect from the database",
								  KeyStroke.getKeyStroke(KeyEvent.VK_D, ActionEvent.ALT_MASK),
								  KeyEvent.VK_D,
								  viewerState, this));
	menu_file.add(menu_file_disconnect);

	/*
	menu_file.addSeparator();

	JMenuItem menu_file_changeconfig = new JMenuItem("Change configuration", new ImageIcon("menu_file_changeconfig.png"));
	menu_file_changeconfig.setMnemonic(KeyEvent.VK_H);
	menu_file_changeconfig.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_H, ActionEvent.ALT_MASK));
	menu_file.add(menu_file_changeconfig);

	JMenuItem menu_file_openconfig = new JMenuItem("Open configuration", new ImageIcon("menu_file_openconfig.png"));
	menu_file_openconfig.setMnemonic(KeyEvent.VK_O);
	menu_file_openconfig.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O, ActionEvent.ALT_MASK));
	menu_file.add(menu_file_openconfig);

	JMenuItem menu_file_saveconfig = new JMenuItem("Save configuration", new ImageIcon("menu_file_saveconfig.png"));
	menu_file_saveconfig.setMnemonic(KeyEvent.VK_S);
	menu_file_saveconfig.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S, ActionEvent.ALT_MASK));
	menu_file.add(menu_file_saveconfig);
	*/

	menu_file.addSeparator();

	JMenuItem menu_file_exit = new JMenuItem(new ActionExit("Exit",	new ImageIcon("menu_file_exit.png"),
								"Exit gnetwatch",
								KeyStroke.getKeyStroke(KeyEvent.VK_X, ActionEvent.ALT_MASK),
								KeyEvent.VK_X));

	menu_file.add(menu_file_exit);

	menu_bar.add(menu_file);

	////////////////////////////////////////////////////////////

	JMenu menu_network = new JMenu("Network");
	menu_network.setMnemonic(KeyEvent.VK_N);

	menu_network_play =
	    new JMenuItem(new ActionPlay("Play",
					 new ImageIcon("menu_network_play.png"),
					 "play events",
					 KeyStroke.getKeyStroke(KeyEvent.VK_P, ActionEvent.ALT_MASK),
					 KeyEvent.VK_P,
					 viewerState, this));
	menu_network.add(menu_network_play);

	menu_network_pause =
	    new JMenuItem(new ActionPlay("Pause",
					 new ImageIcon("menu_network_pause.png"),
					 "stop playing events",
					 KeyStroke.getKeyStroke(KeyEvent.VK_S, ActionEvent.ALT_MASK),
					 KeyEvent.VK_S,
					 viewerState, this));
	menu_network.add(menu_network_pause);

	menu_network.addSeparator();

	menu_network_backward = 
	    new JMenuItem(new ActionBackward("Backward",
					     new ImageIcon("menu_network_backward.png"),
					     "one event back",
					     KeyStroke.getKeyStroke(KeyEvent.VK_B, ActionEvent.ALT_MASK),
					     KeyEvent.VK_B,
					     viewerState, this));
	menu_network.add(menu_network_backward);

	menu_network_forward =
	    new JMenuItem(new ActionForward("Forward",
					    new ImageIcon("menu_network_forward.png"),
					    "one event forward",
					    KeyStroke.getKeyStroke(KeyEvent.VK_F, ActionEvent.ALT_MASK),
					    KeyEvent.VK_F,
					    viewerState, this));
	menu_network.add(menu_network_forward);

	menu_network.addSeparator();

	menu_network_backward_max =
	    new JMenuItem(new ActionBackwardMax("Beginning",
						new ImageIcon("menu_network_backward_max.png"),
						"rewind to first timestamp",
						KeyStroke.getKeyStroke(KeyEvent.VK_G, ActionEvent.ALT_MASK),
						KeyEvent.VK_G,
						viewerState, this));
	menu_network.add(menu_network_backward_max);

	menu_network_forward_max =
	    new JMenuItem(new ActionForwardMax("End",
					       new ImageIcon("menu_network_forward_max.png"),
					       "go to current state",
					       KeyStroke.getKeyStroke(KeyEvent.VK_E, ActionEvent.ALT_MASK),
					       KeyEvent.VK_E,
					       viewerState, this));
	menu_network.add(menu_network_forward_max);

	menu_bar.add(menu_network);

	////////////////////////////////////////////////////////////
	// Menu Windows

	JMenu menu_windows = new JMenu("Windows");
	menu_windows.setMnemonic(KeyEvent.VK_W);

	menu_windows_keys = new JCheckBoxMenuItem(new ActionWinKeys("Keys",
								    KeyStroke.getKeyStroke(KeyEvent.VK_K,
											   ActionEvent.ALT_MASK),
								    KeyEvent.VK_K,
								    viewerState, this));
	menu_windows_keys.setSelected(true);
	menu_windows.add(menu_windows_keys);

	menu_windows_history = new JCheckBoxMenuItem(new ActionWinHistory("History",
									  KeyStroke.getKeyStroke(KeyEvent.VK_H,
												 ActionEvent.ALT_MASK),
									  KeyEvent.VK_H,
									  viewerState, this));
	menu_windows_history.setSelected(true);
	menu_windows.add(menu_windows_history);

	menu_windows_toolbar = new JCheckBoxMenuItem(new ActionWinToolbar("Toolbar",
									  KeyStroke.getKeyStroke(KeyEvent.VK_T,
												 ActionEvent.ALT_MASK),
									  KeyEvent.VK_T,
									  viewerState, this));
	menu_windows_toolbar.setSelected(true);
	menu_windows.add(menu_windows_toolbar);

	menu_windows_info = new JCheckBoxMenuItem(new ActionWinInfo("Info",
								    KeyStroke.getKeyStroke(KeyEvent.VK_I,
											   ActionEvent.ALT_MASK),
								    KeyEvent.VK_I,
								    viewerState, this));
	menu_windows_info.setSelected(true);
	menu_windows.add(menu_windows_info);

	menu_windows.addSeparator();

	menu_windows_startdebug = new JCheckBoxMenuItem("Debug output");
	menu_windows_startdebug.setMnemonic(KeyEvent.VK_D);
	menu_windows_startdebug.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_D, ActionEvent.ALT_MASK));
	menu_windows_startdebug.setState(false);
	menu_windows.add(menu_windows_startdebug);

	menu_windows.addSeparator();

	/*
	ButtonGroup menu_windows_buttongroup = new ButtonGroup();
	
	JRadioButtonMenuItem menu_windows_map = new JRadioButtonMenuItem("Show network map");
	menu_windows_map.setMnemonic(KeyEvent.VK_N);
	menu_windows_map.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_N, ActionEvent.ALT_MASK));
	menu_windows.add(menu_windows_map);
	menu_windows_buttongroup.add(menu_windows_map);

	JRadioButtonMenuItem menu_windows_conf = new JRadioButtonMenuItem("Show configuration");
	menu_windows_conf.setMnemonic(KeyEvent.VK_C);
	menu_windows_conf.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, ActionEvent.ALT_MASK));
	menu_windows.add(menu_windows_conf);
	menu_windows_buttongroup.add(menu_windows_conf);

	JRadioButtonMenuItem menu_windows_debug = new JRadioButtonMenuItem("Show debug");
	menu_windows_debug.setMnemonic(KeyEvent.VK_D);
	menu_windows_debug.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_D, ActionEvent.ALT_MASK));
	menu_windows.add(menu_windows_debug);
	menu_windows_buttongroup.add(menu_windows_debug);

	menu_windows.addSeparator();
	*/

	JMenuItem menu_windows_open = new JMenuItem(new ActionWinOpen("Show all windows",
								      KeyStroke.getKeyStroke(KeyEvent.VK_A,
											     ActionEvent.ALT_MASK),
								      KeyEvent.VK_A,
								      viewerState, this));
	menu_windows_open.setSelected(true);
	menu_windows.add(menu_windows_open);

	JMenuItem menu_windows_close = new JMenuItem(new ActionWinClose("Hide all windows",
									KeyStroke.getKeyStroke(KeyEvent.VK_H,
											       ActionEvent.ALT_MASK),
									KeyEvent.VK_H,
									viewerState, this));
	menu_windows_close.setSelected(true);
	menu_windows.add(menu_windows_close);

	menu_bar.add(menu_windows);

	////////////////////////////////////////////////////////////
	// Menu Debug

	JMenu menu_debug = new JMenu("Debug");
	menu_debug.setMnemonic(KeyEvent.VK_D);

	JMenuItem menu_debug_dumpbdd = new JMenuItem("Dump database");
	menu_debug_dumpbdd.setMnemonic(KeyEvent.VK_D);
	menu_debug_dumpbdd.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_D, ActionEvent.ALT_MASK));
	menu_debug.add(menu_debug_dumpbdd);

	menu_debug_dumpbdd.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    try {
			if (e.getID() == ActionEvent.ACTION_PERFORMED) {
			    addMessage("<P><B>Dumping BDD</B><BR><FONT SIZE=-2><P>" +
				       remoteAccess.DumpDatabase() +
				       "<FONT SIZE=-2>");
			}
		    }

		    catch (RemoteException ex) {
			Debug.logExceptionAndExit(ex);
		    }
		}
	    });

	JMenuItem menu_debug_testbdd = new JMenuItem("Test access to remote database");

	menu_debug_testbdd.setMnemonic(KeyEvent.VK_T);
	menu_debug_testbdd.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_T, ActionEvent.ALT_MASK));
	menu_debug.add(menu_debug_testbdd);

	menu_debug_testbdd.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    try {
			if (e.getID() == ActionEvent.ACTION_PERFORMED) {
			    // remoteAccess.Ping();
			    DebugBdd.remoteAccess.Ping();
			    Debug.debugln("Remote BDD access tested.");
			}
		    }

		    catch (RemoteException ex) {
			Debug.logExceptionAndExit(ex);
		    }
		}
	    });

	menu_bar.add(menu_debug);

	////////////////////////////////////////////////////////////
	// Menu ?

	JMenu menu_help = new JMenu("?");

	JMenuItem menu_help_aboutgnetwatch = new JMenuItem("About gnetwatch", new ImageIcon("menu_aboutgnetwatch.png"));
	menu_help_aboutgnetwatch.setMnemonic(KeyEvent.VK_G);
	menu_help_aboutgnetwatch.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_G, ActionEvent.ALT_MASK));
	menu_help_aboutgnetwatch.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    JOptionPane.showMessageDialog(frame,
						  new JLabel("GNetWatch by Alexandre Fenyo (alex@fenyo.net)"),
						  "About gnetwatch",
						  JOptionPane.ERROR_MESSAGE);
		}
	    });
	menu_help.add(menu_help_aboutgnetwatch);

	JMenuItem menu_help_aboutambience = new JMenuItem("About Ambience", new ImageIcon("menu_aboutambience.png"));
	menu_help_aboutambience.setMnemonic(KeyEvent.VK_A);
	menu_help_aboutambience.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_A, ActionEvent.ALT_MASK));
	menu_help.add(menu_help_aboutambience);
	menu_help_aboutambience.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    JOptionPane.showMessageDialog(frame,
						  new JLabel("Ambience : http://www.extra.research.philips.com/euprojects/ambience"),
						  "About Ambience",
						  JOptionPane.ERROR_MESSAGE);
		}
	    });

	JMenuItem menu_help_aboutenst = new JMenuItem("About Tlcom Paris", new ImageIcon("menu_about.png"));
	menu_help_aboutenst.setMnemonic(KeyEvent.VK_T);
	menu_help_aboutenst.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_T, ActionEvent.ALT_MASK));
	menu_help.add(menu_help_aboutenst);
	menu_help_aboutenst.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    JOptionPane.showMessageDialog(frame,
						  new JLabel("Tlcom Paris : http://www.enst.fr"),
						  "About Tlcom Paris",
						  JOptionPane.ERROR_MESSAGE);
		}
	    });

	menu_help.addSeparator();

	JMenuItem menu_help_help = new JMenuItem("Help", new ImageIcon("menu_help.png"));
	menu_help_help.setMnemonic(KeyEvent.VK_H);
	menu_help_help.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_H, ActionEvent.ALT_MASK));
	menu_help.add(menu_help_help);
	menu_help_help.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    JOptionPane.showMessageDialog(frame,
						  new JLabel("Contact alex@fenyo.net to get support"),
						  "Help",
						  JOptionPane.ERROR_MESSAGE);
		}
	    });

	menu_bar.add(menu_help);

	////////////////////////////////////////////////////////////

	frame.setJMenuBar(menu_bar);

	////////////////////////////////////////////////////////////
	// Toolbar

	toolbar = new JToolBar();
	
	toolbar_file_connect = new JButton(new ActionConnect("",
							     new ImageIcon("toolbar_file_connect.png"),
							     "Connect to a database",
							     KeyStroke.getKeyStroke(KeyEvent.VK_C, ActionEvent.ALT_MASK),
							     KeyEvent.VK_C,
							     viewerState, this));
	toolbar_file_connect.setEnabled(false);
	toolbar.add(toolbar_file_connect);

	toolbar_file_disconnect = new JButton(new ActionDisconnect("",
								   new ImageIcon("toolbar_file_disconnect.png"),
								   "Disconnect from the database",
								   KeyStroke.getKeyStroke(KeyEvent.VK_D, ActionEvent.ALT_MASK),
								   KeyEvent.VK_D,
								   viewerState, this));
	toolbar.add(toolbar_file_disconnect);

	toolbar.addSeparator();

	toolbar_network_pause =
	    new JButton(new ActionPause("",
					new ImageIcon("toolbar_network_pause.png"),
					"stop playing events",
					KeyStroke.getKeyStroke(KeyEvent.VK_S, ActionEvent.ALT_MASK),
					KeyEvent.VK_S,
					viewerState, this));
	toolbar.add(toolbar_network_pause);

 	toolbar_network_backward_max =
	    new JButton(new ActionBackwardMax("",
					      new ImageIcon("toolbar_network_backward_max.png"),
					      "go to first state",
					      KeyStroke.getKeyStroke(KeyEvent.VK_G, ActionEvent.ALT_MASK),
					      KeyEvent.VK_G,
					      viewerState, this));
 	toolbar.add(toolbar_network_backward_max);

	toolbar_network_backward = 
	    new JButton(new ActionBackward("",
					   new ImageIcon("toolbar_network_backward.png"),
					   "one event back",
					   KeyStroke.getKeyStroke(KeyEvent.VK_B, ActionEvent.ALT_MASK),
					   KeyEvent.VK_B,
					   viewerState, this));
	toolbar.add(toolbar_network_backward);


	toolbar_network_play =
	    new JButton(new ActionPlay("",
				       new ImageIcon("toolbar_network_play.png"),
				       "play events",
				       KeyStroke.getKeyStroke(KeyEvent.VK_P, ActionEvent.ALT_MASK),
				       KeyEvent.VK_P,
				       viewerState, this));
	toolbar.add(toolbar_network_play);

	toolbar_network_forward =
 	    new JButton(new ActionForward("",
					  new ImageIcon("toolbar_network_forward.png"),
					  "one event forward",
					  KeyStroke.getKeyStroke(KeyEvent.VK_F, ActionEvent.ALT_MASK),
					  KeyEvent.VK_F,
					  viewerState, this));
	toolbar.add(toolbar_network_forward);

	toolbar_network_forward_max =
	    new JButton(new ActionForwardMax("",
					      new ImageIcon("toolbar_network_forward_max.png"),
					      "go to current state",
					      KeyStroke.getKeyStroke(KeyEvent.VK_E, ActionEvent.ALT_MASK),
					      KeyEvent.VK_E,
					      viewerState, this));
	toolbar.add(toolbar_network_forward_max);

	frame.getContentPane().add(toolbar, BorderLayout.NORTH);

	////////////////////////////////////////////////////////////
	// build view
	view = new View("background.png");
	DebugGUI.view = view;
	view.setLayout(null);

	// build scrollPane
	JScrollPane scrollPane = new JScrollPane(view);
	DebugGUI.scrollPane = scrollPane;
	scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
	scrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);

	/*
	// build vpict
	vpict = new VPict(view, "robot.png");
	vpict.setLocationAndInform(600, 600);
	DebugGUI.vpict = vpict;

	// for testing
	VPict robot = new VPict(view, "robot.png");
	robot.setLocationAndInform(500, 500);

	// for testing
	VConnector connector = new VConnector(view, robot, vpict);

	// add graphic objects
	view.add(new VText(view, vpict, "vpict"));
	view.add(new VText(view, connector, "connector"));
	view.add(new VText(view, robot, "robot"));
	view.add(connector);
	view.add(robot);
	view.add(vpict);
	*/

	////////////////////////////////////////////////////////////
	// Debug pane

	debugText = new JEditorPane();
	debugText.setEditable(false);
	debugText.setContentType("text/html");

	addMessage("<B>Starting gnetwatch</B>");

	JScrollPane debugScrollPane = new JScrollPane(debugText);
	debugScrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);

	////////////////////////////////////////////////////////////
	// Tabbed pane

	JTabbedPane tabbedPane = new JTabbedPane();
	tabbedPane.addTab("<HTML><I>Network Map</I></HTML>", null, scrollPane);
	tabbedPane.setToolTipTextAt(0, "Network map");

	tabbedPane.addTab("<HTML><I>Messages</I></HTML>", null, debugScrollPane);
	tabbedPane.setToolTipTextAt(1, "Information and debug messages");

	frame.getContentPane().add(tabbedPane, BorderLayout.CENTER);

	////////////////////////////////////////////////////////////

	JLabel infoLabel = new JLabel(" gnetwatch by A. Fenyo - 2002 ");
	infoLabel.setForeground(new Color(156 - 20, 154 - 20, 190 - 20));
	infoLabel.setBorder(BorderFactory.createEtchedBorder());
	timestampLabel = new JLabel("<HTML>&nbsp;<FONT COLOR=red>waiting for database update</FONT>&nbsp;</HTML>");
	timestampLabel.setBorder(BorderFactory.createEtchedBorder());
	timestampLabel.setForeground(new Color(156 - 20, 154 - 20, 190 - 20));

	BorderLayout bottomLayout = new BorderLayout();
	bottomPanel = new JPanel(bottomLayout);
	bottomPanel.add(infoLabel, BorderLayout.CENTER);
	bottomPanel.add(timestampLabel, BorderLayout.EAST);

	frame.getContentPane().add(bottomPanel, BorderLayout.SOUTH);

	////////////////////////////////////////////////////////////

	DefaultMutableTreeNode topTreeNode = new DefaultMutableTreeNode("<HTML><FONT COLOR=red><B>waiting for update</B></FONT></HTML>");
	TreeModel treeModel = new DefaultTreeModel(topTreeNode);

	keysTree = new JTree(treeModel);
	DefaultTreeCellRenderer treeCellRenderer = new DefaultTreeCellRenderer();
	treeCellRenderer.setBackgroundNonSelectionColor(new Color(0xFBFEDA));
	keysTree.setCellRenderer(treeCellRenderer);
	keysTree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
	keysTree.setEditable(false);

	keysTree.addTreeSelectionListener(new TreeSelectionListener() {
		public void valueChanged(TreeSelectionEvent e) {
		    DefaultMutableTreeNode node = (DefaultMutableTreeNode) keysTree.getLastSelectedPathComponent();
		    if (node == null) return;
		    Object nodeInfo = node.getUserObject();
		    if (node.isLeaf() && IdKeyRefNode.class.isInstance(nodeInfo)) {
			Collection history = null;
			IdKeyRefNode idKeyRefNode = (IdKeyRefNode) nodeInfo;

			try {
			    history = remoteAccess.GetHistory(idKeyRefNode.identity,
							      idKeyRefNode.key,
							      idKeyRefNode.reference);
			}
			catch (RemoteException ex) {
			    Debug.logExceptionAndExit(ex);
			}

			displayHistory(history);
		    } else displayHistory(null);
		}
	    });

	JScrollPane keysScrollPane = new JScrollPane(keysTree);
	keysScrollPane.setBorder(BorderFactory.createTitledBorder("Keys"));
	keysScrollPane.setPreferredSize(new Dimension(200, 10));

	Object[][] table_content = { { "none", "none" } };
	String[] table_columns = { "<HTML><B>timestamp</B></HTML>", "<HTML><B>value</B></HTML>" };
	historyTable = new JTable(new DefaultTableModel (table_content, table_columns) {
		public boolean isCellEditable(int row, int col) {
		    return false;
		}
	    });

	historyTable.setAutoResizeMode(JTable.AUTO_RESIZE_ALL_COLUMNS);

	JScrollPane tableScrollPane = new JScrollPane(historyTable);
	tableScrollPane.setBorder(BorderFactory.createTitledBorder("History"));
	tableScrollPane.setPreferredSize(new Dimension(180, 10));
	//	tableScrollPane.getViewport().setBackground(Color.lightGray);

 	historyTable.getParent().getParent().setBackground(new Color(0xFBFEDA));
 	keysTree.getParent().getParent().setBackground(new Color(0xFBFEDA));

	historyTable.getParent().setBackground(historyTable.getParent().getParent().getBackground());
	keysTree.getParent().setBackground(keysTree.getParent().getParent().getBackground());
	historyTable.setBackground(historyTable.getParent().getBackground());
	keysTree.setBackground(keysTree.getParent().getBackground());

	splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, keysScrollPane, tableScrollPane);
	splitPane.setBorder(BorderFactory.createMatteBorder(2, 2, 2, 2, Color.black));
	frame.getContentPane().add(splitPane, BorderLayout.WEST);
	//	splitPane.addPropertyChangeListener("resizeWeight", new PropertyChangeListener() {
	splitPane.addPropertyChangeListener(JSplitPane.DIVIDER_LOCATION_PROPERTY, new PropertyChangeListener() {
		public void propertyChange(PropertyChangeEvent e) {

		    if (splitPane.getDividerLocation() > splitPane.getMaximumDividerLocation()) {
			menu_windows_keys.setState(true);
			menu_windows_history.setState(false);
		    } else {
			if (splitPane.getDividerLocation() < splitPane.getMinimumDividerLocation()) {
			    menu_windows_keys.setState(false);
			    menu_windows_history.setState(true);
			} else {
			    menu_windows_keys.setState(true);
			    menu_windows_history.setState(true);
			}
		    }
		}
	    });

	////////////////////////////////////////////////////////////

	setEnabledEnd();

	////////////////////////////////////////////////////////////

	// do the whole job for the GUI
 	frame.setUndecorated(true);
 	frame.getRootPane().setWindowDecorationStyle(JRootPane.FRAME);
	frame.pack();

	Dimension screen_dim = frame.getToolkit().getScreenSize();
	frame.setLocation((screen_dim.width - initial_width) / 2, (screen_dim.height - initial_height) / 2);
	frame.setSize(initial_width, initial_height);

	view.revalidate();

	// frame.setResizable(false);

	// frame.show();
	// frame.setVisible(false);

	Viewer.setTextIntro("Building GUI objects");

	////////////////////////////////////////////////////////////

	int cpt;
	for (cpt = 1; cpt < 50; cpt++) {
	    try { Thread.currentThread().sleep(100); }
	    catch (InterruptedException ex) { Debug.logException(ex); }
	    Viewer.setTextIntroImmediate("Building GUI objects (" + new Integer(cpt * 2).toString() + "%)");
	}

	Viewer.setTextIntro("gnetwatch by fenyo@enst.fr");
 	try { Thread.currentThread().sleep(2000); }
 	catch (InterruptedException ex) { Debug.logException(ex); }
	    
	introFrame.dispose();

	frame.show();
	// frame.setVisible(true);
    }

    private String message = "";
    public void addMessage(String msg) {
	msg = msg.replaceAll("\\n", "<BR>");
	msg = msg.replaceAll("\\r", "<BR>");
	message += msg;
	debugText.setText(message);
    }

    // access by main thread
    public Thread startThread(RemoteAccess remoteAccess) {
	DebugBdd.remoteAccess = this.remoteAccess = remoteAccess;
       	viewerState = new ViewerState(remoteAccess);
	viewerDisplay = new ViewerDisplay(this);
	thr = new Thread(this, "Viewer");
	BuildGUI();
	thr.start();
	return thr;
    }

    public Thread restartThread(RemoteAccess remoteAccess) {
	DebugBdd.remoteAccess = this.remoteAccess = remoteAccess;
       	viewerState.Reconnect(remoteAccess);
        viewerDisplay.informEnd();
	viewerDisplay = new ViewerDisplay(this);
	setEnabledReconnected();

	thr = new Thread(this, "Viewer");
	thr.start();
	return thr;
    }

    public void stopThread() {
	if (thr != null) {
	    thr.interrupt();

	    try {
		thr.join();
	    }
	    catch (InterruptedException ex) {
		Debug.logException(this, ex);
	    }

	    try {
		remoteAccess.Disconnect(viewerState.remoteId);
	    }
	    catch (RemoteException ex) {
		Debug.logException(this, ex);
	    }

	    viewerState.remoteAccess = null;
	    setEnabledDisconnected();
	}
    }

    public static void loadIntro() {
	introFrame = new JFrame("gnetwatch - loading");

	IntroPanel introPanel = new IntroPanel("intro.png");
	introPanel.setLayout(null);

	introLabel = new JLabel("");

	introPanel.add(introLabel);

	introFrame.getContentPane().add(introPanel);
	introFrame.setUndecorated(true);
	introFrame.pack();

	Dimension image_dim = introPanel.getDimension();
	Dimension screen_dim = introFrame.getToolkit().getScreenSize();

	introLabel.setBounds(345, 48, 500, 50);

	introFrame.setSize(image_dim);
	introFrame.setLocation((screen_dim.width - image_dim.width) / 2, (screen_dim.height - image_dim.height) / 2);
	introPanel.revalidate();

	introFrame.show();
    }

    public static void setTextIntro(final String str) {
	if (introLabel != null) try {
	    SwingUtilities.invokeAndWait(new Runnable () { 
		    public void run() {
			try { Thread.currentThread().sleep(300); }
			catch (InterruptedException ex) { Debug.logException(ex); }
			introLabel.setText("");
		    } });
	    SwingUtilities.invokeAndWait(new Runnable () { 
		    public void run() {
			// wait some time for the new text to be read
			try { Thread.currentThread().sleep(100); }
			catch (InterruptedException ex) { Debug.logException(ex); }
			introLabel.setText(str);
		    } });
	}
	catch (InterruptedException ex) {
	    Debug.logExceptionAndExit(ex);
	}
	catch (InvocationTargetException ex) {
	    Debug.logExceptionAndExit(ex);
	}
    }

    public static void setTextIntroImmediate(final String str) {
	if (introLabel != null) try {
	    SwingUtilities.invokeAndWait(new Runnable () { 
		    public void run() {
			introLabel.setText(str);
		    } });
	}
	catch (InterruptedException ex) {
	    Debug.logExceptionAndExit(ex);
	}
	catch (InvocationTargetException ex) {
	    Debug.logExceptionAndExit(ex);
	}
    }

    public void displayDate(Date date) {
	if (date.compareTo(new Date(Long.MAX_VALUE)) == 0)
	    timestampLabel.setText("<HTML>&nbsp;current state&nbsp;</HTML>");
	else if (date.compareTo(new Date(Long.MIN_VALUE)) == 0)
	    timestampLabel.setText("<HTML>&nbsp;rewind&nbsp;</HTML>");
	else timestampLabel.setText("<HTML>&nbsp;" + date.toString() + "&nbsp;</HTML>");
    }

    public void displayKeysTree(State state) {
	String labeldate;

	synchronized (state) {
	    DefaultMutableTreeNode topTreeNode = (DefaultMutableTreeNode) keysTree.getModel().getRoot();

	    // update content of top tree node (current date)
	    if (state.date.compareTo(new Date(Long.MAX_VALUE)) == 0)
		labeldate = "<HTML><B>current state</B></HTML>";
	    else if (state.date.compareTo(new Date(Long.MIN_VALUE)) == 0)
		labeldate = "<HTML><B>rewind</B></HTML>";
	    else labeldate = "<HTML><B>" + state.date.toString() + "</B></HTML>";

	    topTreeNode.setUserObject(labeldate);

	    // fill in the tree
	    topTreeNode.removeAllChildren();
	    HashMap idref2node = new HashMap();
	    HashMap id2node = new HashMap();
	    EntryKey map_key;
	    for (Iterator it = state.map.keySet().iterator(); it.hasNext();) {
		map_key = (EntryKey) it.next();
		//		dump_str += "  - ([id=" + map_key.getIdentity() + "], [key=" + map_key.getKey() + "])\n";

		DefaultMutableTreeNode node_identity;
		if (id2node.containsKey(map_key.getIdentity()))
		    node_identity = (DefaultMutableTreeNode) id2node.get(map_key.getIdentity());
		else {
		    node_identity = new DefaultMutableTreeNode("<HTML><B>" +
							       map_key.getIdentity() +
							       "</B></HTML>");
		    id2node.put(map_key.getIdentity(), node_identity);
		    topTreeNode.add(node_identity);
		}

		String ref;
		TreeMap reflist = (TreeMap) state.map.get(map_key);
		synchronized (reflist) {
		    for (Iterator it2 = reflist.keySet().iterator(); it2.hasNext();) {
			ref = (String) it2.next();

			EntryVal entryval = (EntryVal) reflist.get(ref);

			if (ref.equals("")) {
			    DefaultMutableTreeNode node_key =
				new DefaultMutableTreeNode(new IdKeyRefNode("<HTML>" +
									    map_key.getKey() + " : " + entryval.getValue() +
									    "</HTML>",
									    map_key.getIdentity(), map_key.getKey(), ""));

			    node_identity.add(node_key);

			} else {

			    DefaultMutableTreeNode node_idref = null;
			    if (idref2node.containsKey(map_key.getIdentity() + "/" + ref))
				node_idref = (DefaultMutableTreeNode)
				    idref2node.get(map_key.getIdentity() + "/" + ref);
			    else {
				node_idref =
				    new DefaultMutableTreeNode("<HTML>" + ref + "</HTML>");
				idref2node.put(map_key.getIdentity() + "/" + ref, node_idref);
				node_identity.add(node_idref);
			    }

			    DefaultMutableTreeNode node_key =
				new DefaultMutableTreeNode(new IdKeyRefNode("<HTML>" +
									    map_key.getKey() + " : " + entryval.getValue() +
									    "</HTML>",
									    map_key.getIdentity(), map_key.getKey(), ref));

			    node_idref.add(node_key);
			}
		    }
		}
	    }

	    ((DefaultTreeModel) keysTree.getModel()).reload();
	}
    }

    public void displayHistory(Collection history) {
	String[] table_columns = { "<HTML><B>timestamp</B></HTML>", "<HTML><B>value</B></HTML>" };

	if (history == null || history.size() == 0) {
	    Object[][] table_content = { { "none", "none" } };
	    historyTable.setModel(new DefaultTableModel (table_content, table_columns) {
		    public boolean isCellEditable(int row, int col) {
			return false;
		    }
		});

	} else {

	    String table_content[][] = new String [history.size()][2];

	    Date current;
	    synchronized (viewerState.getState()) {
		current = (Date) viewerState.getState().date.clone();
	    }

	    for (int i = 0; i < history.size(); i++) {
		if (((EntryVal) (history.toArray()[i])).getTimestamp().equals(current))
		    table_content[i][0] =
			"[" + ((EntryVal) (history.toArray()[i])).getTimestamp().toString() + "]";
		else table_content[i][0] = ((EntryVal) (history.toArray()[i])).getTimestamp().toString();
		table_content[i][1] = ((EntryVal) (history.toArray()[i])).getValue();
	    }

	    historyTable.setModel(new DefaultTableModel (table_content, table_columns) {
		    public boolean isCellEditable(int row, int col) {
			return false;
		    }
		});
	}
    }

    public void setEnabledBegin() {
	toolbar_network_backward.setEnabled(false);
	toolbar_network_backward_max.setEnabled(false);
	toolbar_network_forward.setEnabled(true);
	toolbar_network_forward_max.setEnabled(true);
	toolbar_network_play.setEnabled(true);
	toolbar_network_pause.setEnabled(false);

	menu_network_backward.setEnabled(false);
	menu_network_backward_max.setEnabled(false);
	menu_network_forward.setEnabled(true);
	menu_network_forward_max.setEnabled(true);
	menu_network_play.setEnabled(true);
	menu_network_pause.setEnabled(false);
    }

    public void setEnabledEnd() {
	toolbar_network_backward.setEnabled(true);
	toolbar_network_backward_max.setEnabled(true);
	toolbar_network_forward.setEnabled(false);
	toolbar_network_forward_max.setEnabled(false);
	toolbar_network_play.setEnabled(false);
	toolbar_network_pause.setEnabled(false);

	menu_network_backward.setEnabled(true);
	menu_network_backward_max.setEnabled(true);
	menu_network_forward.setEnabled(false);
	menu_network_forward_max.setEnabled(false);
	menu_network_play.setEnabled(false);
	menu_network_pause.setEnabled(false);
    }

    public void setEnabledPlay() {
	toolbar_network_backward.setEnabled(true);
	toolbar_network_backward_max.setEnabled(true);
	toolbar_network_forward.setEnabled(true);
	toolbar_network_forward_max.setEnabled(true);
	toolbar_network_play.setEnabled(false);
	toolbar_network_pause.setEnabled(true);

	menu_network_backward.setEnabled(true);
	menu_network_backward_max.setEnabled(true);
	menu_network_forward.setEnabled(true);
	menu_network_forward_max.setEnabled(true);
	menu_network_play.setEnabled(false);
	menu_network_pause.setEnabled(true);
    }

    public void setEnabledPause() {
	toolbar_network_backward.setEnabled(true);
	toolbar_network_backward_max.setEnabled(true);
	toolbar_network_forward.setEnabled(true);
	toolbar_network_forward_max.setEnabled(true);
	toolbar_network_play.setEnabled(true);
	toolbar_network_pause.setEnabled(false);

	menu_network_backward.setEnabled(true);
	menu_network_backward_max.setEnabled(true);
	menu_network_forward.setEnabled(true);
	menu_network_forward_max.setEnabled(true);
	menu_network_play.setEnabled(true);
	menu_network_pause.setEnabled(false);
    }

    public void setEnabledDisconnected() {
	toolbar_network_backward.setEnabled(false);
	toolbar_network_backward_max.setEnabled(false);
	toolbar_network_forward.setEnabled(false);
	toolbar_network_forward_max.setEnabled(false);
	toolbar_network_play.setEnabled(false);
	toolbar_network_pause.setEnabled(false);

	menu_network_backward.setEnabled(false);
	menu_network_backward_max.setEnabled(false);
	menu_network_forward.setEnabled(false);
	menu_network_forward_max.setEnabled(false);
	menu_network_play.setEnabled(false);
	menu_network_pause.setEnabled(false);

	toolbar_file_connect.setEnabled(true);
	menu_file_connect.setEnabled(true);
	toolbar_file_disconnect.setEnabled(false);
	menu_file_disconnect.setEnabled(false);
    }

    public void setEnabledReconnected() {
	toolbar_file_connect.setEnabled(false);
	menu_file_connect.setEnabled(false);
	toolbar_file_disconnect.setEnabled(true);
	menu_file_disconnect.setEnabled(true);
    }
}

////////////////////////////////////////////////////////////

abstract class GeneralAction extends AbstractAction {
    public GeneralAction(String text, ImageIcon icon, String desc, KeyStroke accel, int mnemonic) {
	super(text, icon);
	putValue(SHORT_DESCRIPTION, desc);
	putValue(ACCELERATOR_KEY, accel);
	putValue(MNEMONIC_KEY, new Integer(mnemonic));
    }
}

abstract class SimpleAction extends AbstractAction {
    public SimpleAction(String text, KeyStroke accel, int mnemonic) {
	super(text);
	putValue(ACCELERATOR_KEY, accel);
	putValue(MNEMONIC_KEY, new Integer(mnemonic));
    }
}

class ActionWinKeys extends SimpleAction {
    ViewerState viewerState;
    Viewer viewer;

    public ActionWinKeys(String text, KeyStroke accel, int mnemonic,
			 ViewerState viewerState, Viewer viewer) {
	super(text, accel, mnemonic);
	this.viewerState = viewerState;
	this.viewer = viewer;
    }

    public void actionPerformed(ActionEvent e) {
	if (viewer.splitPane_removed == true) {
	    viewer.splitPane.setDividerLocation(1.0);
	    viewer.frame.getContentPane().add(viewer.splitPane, BorderLayout.WEST);
	    viewer.frame.getContentPane().validate();
	    viewer.splitPane_removed = false;
	    return;
	}

	if (viewer.menu_windows_keys.getState() == true) {
	    if (viewer.menu_windows_history.getState() == true)
		viewer.splitPane.setDividerLocation(0.5);
	    else viewer.splitPane.setDividerLocation(1.0);
	}
	else {
	    if (viewer.menu_windows_history.getState() == false) {
		viewer.frame.getContentPane().remove(viewer.splitPane);
		viewer.frame.getContentPane().validate();
		viewer.splitPane_removed = true;
	    } else {
		viewer.splitPane.setDividerLocation(0);
	    }

	}
    }
}

class ActionWinHistory extends SimpleAction {
    ViewerState viewerState;
    Viewer viewer;

    public ActionWinHistory(String text, KeyStroke accel, int mnemonic,
			    ViewerState viewerState, Viewer viewer) {
	super(text, accel, mnemonic);
	this.viewerState = viewerState;
	this.viewer = viewer;
    }

    public void actionPerformed(ActionEvent e) {
	if (viewer.splitPane_removed == true) {
	    viewer.splitPane.setDividerLocation(0);
	    viewer.frame.getContentPane().add(viewer.splitPane, BorderLayout.WEST);
	    viewer.frame.getContentPane().validate();
	    viewer.splitPane_removed = false;
	    return;
	}

	if (viewer.menu_windows_history.getState() == true) {
	    if (viewer.menu_windows_keys.getState() == true)
		viewer.splitPane.setDividerLocation(0.5);
	    else viewer.splitPane.setDividerLocation(0);
	}
	else {
	    if (viewer.menu_windows_keys.getState() == false) {
		viewer.frame.getContentPane().remove(viewer.splitPane);
		viewer.frame.getContentPane().validate();
		viewer.splitPane_removed = true;
	    } else {
		viewer.splitPane.setDividerLocation(1.0);
	    }

	}
    }
}

class ActionWinOpen extends SimpleAction {
    ViewerState viewerState;
    Viewer viewer;

    public ActionWinOpen(String text, KeyStroke accel, int mnemonic,
			 ViewerState viewerState, Viewer viewer) {
	super(text, accel, mnemonic);
	this.viewerState = viewerState;
	this.viewer = viewer;
    }

    public void actionPerformed(ActionEvent e) {
	viewer.menu_windows_keys.setState(true);
	viewer.menu_windows_history.setState(true);
	viewer.splitPane.setDividerLocation(0.5);

	if (viewer.splitPane_removed == true) {
	    viewer.frame.getContentPane().add(viewer.splitPane, BorderLayout.WEST);
	    viewer.splitPane_removed = false;
	}

	if (viewer.menu_windows_toolbar.getState() == false) {
	    viewer.frame.getContentPane().add(viewer.toolbar, BorderLayout.NORTH);
	    viewer.menu_windows_toolbar.setState(true);
	}

	if (viewer.menu_windows_info.getState() == false) {
	    viewer.frame.getContentPane().add(viewer.bottomPanel, BorderLayout.SOUTH);
	    viewer.menu_windows_info.setState(true);
	}

	viewer.frame.getContentPane().validate();
    }
}

class ActionWinClose extends SimpleAction {
    ViewerState viewerState;
    Viewer viewer;

    public ActionWinClose(String text, KeyStroke accel, int mnemonic,
			 ViewerState viewerState, Viewer viewer) {
	super(text, accel, mnemonic);
	this.viewerState = viewerState;
	this.viewer = viewer;
    }

    public void actionPerformed(ActionEvent e) {
	viewer.menu_windows_keys.setState(false);
	viewer.menu_windows_history.setState(false);

	if (viewer.splitPane_removed == false) {
	    viewer.frame.getContentPane().remove(viewer.splitPane);
	    viewer.splitPane_removed = true;
	}

	if (viewer.menu_windows_toolbar.getState() == true) {
	    viewer.frame.getContentPane().remove(viewer.toolbar);
	    viewer.menu_windows_toolbar.setState(false);
	}

	if (viewer.menu_windows_info.getState() == true) {
	    viewer.frame.getContentPane().remove(viewer.bottomPanel);
	    viewer.menu_windows_info.setState(false);
	}

	viewer.frame.getContentPane().validate();
    }
}

class ActionWinToolbar extends SimpleAction {
    ViewerState viewerState;
    Viewer viewer;

    public ActionWinToolbar(String text, KeyStroke accel, int mnemonic,
			    ViewerState viewerState, Viewer viewer) {
	super(text, accel, mnemonic);
	this.viewerState = viewerState;
	this.viewer = viewer;
    }

    public void actionPerformed(ActionEvent e) {
	if (viewer.menu_windows_toolbar.getState() == true)
	    viewer.frame.getContentPane().add(viewer.toolbar, BorderLayout.NORTH);
       	else viewer.frame.getContentPane().remove(viewer.toolbar);
	viewer.frame.getContentPane().validate();
    }
}

class ActionWinInfo extends SimpleAction {
    ViewerState viewerState;
    Viewer viewer;

    public ActionWinInfo(String text, KeyStroke accel, int mnemonic,
			 ViewerState viewerState, Viewer viewer) {
	super(text, accel, mnemonic);
	this.viewerState = viewerState;
	this.viewer = viewer;
    }

    public void actionPerformed(ActionEvent e) {
	if (viewer.menu_windows_info.getState() == true)
	    viewer.frame.getContentPane().add(viewer.bottomPanel, BorderLayout.SOUTH);
       	else viewer.frame.getContentPane().remove(viewer.bottomPanel);
	viewer.frame.getContentPane().validate();
    }
}

class ActionConnect extends GeneralAction {
    ViewerState viewerState;
    Viewer viewer;

    public ActionConnect(String text, ImageIcon icon, String desc, KeyStroke accel, int mnemonic,
			 ViewerState viewerState, Viewer viewer) {
	super(text, icon, desc, accel, mnemonic);
	this.viewerState = viewerState;
	this.viewer = viewer;
    }

    public void actionPerformed(ActionEvent e) {
	if (viewerState.remoteAccess != null) return;

	String server_name = (String) JOptionPane.showInputDialog(viewer.frame,
								  new JLabel("database server name ?"),
								  "new connection",
								  JOptionPane.QUESTION_MESSAGE,
								  null, null, "localhost");
	if (server_name == null) return;

	RemoteAccess remoteAccess = null;

	try {
	    remoteAccess = (RemoteAccess)
		Naming.lookup("//" + server_name + ":1099" + "/" + RemoteAccess.LOOKUPNAME);
	}
	catch (MalformedURLException ex) {
	    Debug.logExceptionAndExit(ex);
	}
	catch (NotBoundException ex) {
	    Debug.logException(ex);
	    Debug.errorln("Can't find a database server - no remote access to the database");
	    remoteAccess = null;
	}
	catch (RemoteException ex) {
	    Debug.logException(ex);
	    Debug.errorln("Can't connect to the RMI registry - no remote access to the database");
	    remoteAccess = null;
	}

	if (remoteAccess != null) {
	    viewer.setEnabledEnd();
	    viewer.restartThread(remoteAccess);
	} else JOptionPane.showMessageDialog(viewer.frame,
					     new JLabel("can't connect to the RMI server"),
					     "new connection",
					     JOptionPane.ERROR_MESSAGE);
    }
}

class ActionDisconnect extends GeneralAction {
    ViewerState viewerState;
    Viewer viewer;

    public ActionDisconnect(String text, ImageIcon icon, String desc, KeyStroke accel, int mnemonic,
			    ViewerState viewerState, Viewer viewer) {
	super(text, icon, desc, accel, mnemonic);
	this.viewerState = viewerState;
	this.viewer = viewer;
    }

    public void actionPerformed(ActionEvent e) {
	viewer.stopThread();
    }
}

class ActionExit extends GeneralAction {
    public ActionExit(String text, ImageIcon icon, String desc, KeyStroke accel, int mnemonic) {
	super(text, icon, desc, accel, mnemonic);
    }

    public void actionPerformed(ActionEvent e) {
	System.exit(0);
    }
}

class ActionPlay extends GeneralAction {
    ViewerState viewerState;
    Viewer viewer;

    public ActionPlay(String text, ImageIcon icon, String desc, KeyStroke accel, int mnemonic,
		      ViewerState viewerState, Viewer viewer) {
	super(text, icon, desc, accel, mnemonic);
	this.viewerState = viewerState;
	this.viewer = viewer;
    }

    public void actionPerformed(ActionEvent e) {
	if (viewer.playing_timer != null) return;

	if (new Date(Long.MAX_VALUE).equals(viewerState.NextDate())) {
	    viewer.setEnabledEnd();
	    return;
	}
	else viewer.setEnabledPlay();

	Action actionPlayTimer = new AbstractAction() {
		public void actionPerformed(ActionEvent e) {
		    if (new Date(Long.MAX_VALUE).equals(viewerState.NextDate())) {
			viewer.playing_timer.stop();
			viewer.playing_timer = null;
			viewer.setEnabledEnd();
		    }
		    else viewer.setEnabledPlay();
		}
	    };

	viewer.playing_timer = new javax.swing.Timer(1000, actionPlayTimer);
	viewer.playing_timer.start();
    }
}

class ActionPause extends GeneralAction {
    ViewerState viewerState;
    Viewer viewer;

    public ActionPause(String text, ImageIcon icon, String desc, KeyStroke accel, int mnemonic,
		       ViewerState viewerState, Viewer viewer) {
	super(text, icon, desc, accel, mnemonic);
	this.viewerState = viewerState;
	this.viewer = viewer;
    }

    public void actionPerformed(ActionEvent e) {
	if (viewer.playing_timer == null) return;

	viewer.playing_timer.stop();
	viewer.playing_timer = null;
	viewer.setEnabledPause();
    }
}

class ActionForwardMax extends GeneralAction {
    ViewerState viewerState;
    Viewer viewer;

    public ActionForwardMax(String text, ImageIcon icon, String desc, KeyStroke accel, int mnemonic,
			    ViewerState viewerState, Viewer viewer) {
	super(text, icon, desc, accel, mnemonic);
	this.viewerState = viewerState;
	this.viewer = viewer;
    }

    public void actionPerformed(ActionEvent e) {
	if (viewer.playing_timer != null) {
	    viewer.playing_timer.stop();
	    viewer.playing_timer = null;
	}
	viewerState.SetDate(new Date(Long.MAX_VALUE));
	viewer.setEnabledEnd();
    }
}

class ActionBackwardMax extends GeneralAction {
    ViewerState viewerState;
    Viewer viewer;

    public ActionBackwardMax(String text, ImageIcon icon, String desc, KeyStroke accel, int mnemonic,
			     ViewerState viewerState, Viewer viewer) {
	super(text, icon, desc, accel, mnemonic);
	this.viewerState = viewerState;
	this.viewer = viewer;
    }

    public void actionPerformed(ActionEvent e) {
	if (viewer.playing_timer != null) {
	    viewer.playing_timer.stop();
	    viewer.playing_timer = null;
	}
	viewerState.SetDate(new Date(Long.MIN_VALUE));
	viewer.setEnabledBegin();
    }
}

class ActionForward extends GeneralAction {
    ViewerState viewerState;
    Viewer viewer;

    public ActionForward(String text, ImageIcon icon, String desc, KeyStroke accel, int mnemonic,
			 ViewerState viewerState, Viewer viewer) {
	super(text, icon, desc, accel, mnemonic);
	this.viewerState = viewerState;
	this.viewer = viewer;
    }

    public void actionPerformed(ActionEvent e) {
	if (viewer.playing_timer != null) {
	    viewer.playing_timer.stop();
	    viewer.playing_timer = null;
	}

	Date nextdate = viewerState.NextDate();
	if (new Date(Long.MAX_VALUE).equals(nextdate))
	    viewer.setEnabledEnd();
	else viewer.setEnabledPause();
    }
}

class ActionBackward extends GeneralAction {
    ViewerState viewerState;
    Viewer viewer;

    public ActionBackward(String text, ImageIcon icon, String desc, KeyStroke accel, int mnemonic,
			  ViewerState viewerState, Viewer viewer) {
	super(text, icon, desc, accel, mnemonic);
	this.viewerState = viewerState;
	this.viewer = viewer;
    }

    public void actionPerformed(ActionEvent e) {
	if (viewer.playing_timer != null) {
	    viewer.playing_timer.stop();
	    viewer.playing_timer = null;
	}

	Date prevdate = viewerState.PrevDate();
	if (new Date(Long.MIN_VALUE).equals(prevdate))
	    viewer.setEnabledBegin();
	else viewer.setEnabledPause();
    }
}

////////////////////////////////////////////////////////////

public class BddManager {
    private static final LogFiles logFiles = new LogFiles(); // access by main & Client threads
    private static final AcceptClients acceptClients = new AcceptClients(); // access by main thread
    private static final Bdd bdd = new Bdd(); // access by main, AcceptClients & Client threads

    // access by main thread
    public static void main(String [] args)
    {
	Conf.setParameters();
	Debug.debugln(BddManager.class, "BddManager started");

	////////////////////////////////////////////////////////////
	// Build "loading" frame

	if (Conf.getMode().equals("server") == false) {
	    Viewer.loadIntro();
	    try { Thread.currentThread().sleep(3000); }
	    catch (InterruptedException ex) { Debug.logException(ex); }
	}

	////////////////////////////////////////////////////////////

	// System.getProperties().list(System.out);

	if (Conf.getMode().equals("server") || Conf.getMode().equals("both")) {
	    Viewer.setTextIntro("Binding BDD server to RMI registry");
	    try {
		Naming.bind(RemoteAccess.LOOKUPNAME, new RemoteAccessImpl(bdd));
	    }
	    catch (MalformedURLException ex) {
		Debug.logExceptionAndExit(ex);
	    }
	    catch (RemoteException ex) {
		Debug.logException(ex);
		Debug.errorln("Can't bind to the RMI registry - do not offer remote access to the database");
	    }
	    catch (AlreadyBoundException ex) {
		Debug.logException(ex);
		Debug.errorln("another server is already bound to the RMI registry - do not offer remote access to the database");
	    }
	}
	
	RemoteAccess remoteAccess = null;
	if (Conf.getMode().equals("client") || Conf.getMode().equals("both")) {
	    Viewer.setTextIntro("Looking for a BDD server");
	    try {
		remoteAccess = (RemoteAccess)
		    Naming.lookup(Conf.getRmiRegistry() + "/" + RemoteAccess.LOOKUPNAME);
	    }
	    catch (MalformedURLException ex) {
		Debug.logExceptionAndExit(ex);
	    }
	    catch (NotBoundException ex) {
		Debug.logException(ex);
		Debug.errorln("Can't find a database server - no remote access to the database");
		remoteAccess = null;
	    }
	    catch (RemoteException ex) {
		Debug.logException(ex);
		Debug.errorln("Can't connect to the RMI registry - no remote access to the database");
		remoteAccess = null;
	    }

	    if (remoteAccess == null) {
		try {
		    remoteAccess = new RemoteAccessImpl(bdd);
		    Debug.debugln("creating a local wrapper to the database");
		}
		catch (RemoteException ex) {
		    Debug.logExceptionAndExit(ex);
		}
	    }
	}

	if (Conf.getMode().equals("server") || Conf.getMode().equals("both")) {
	    Viewer.setTextIntro("Loading initial BDD");

	    String initial_log = logFiles.loadLogFile("bdd.log");
	    logFiles.openLogFile("bdd.log");

	    Viewer.setTextIntro("Start accepting clients");

	    Debug.debugln(BddManager.class, "starting AcceptClients thread");
	    acceptClients.startThread(bdd, logFiles, initial_log);
	}

	if (Conf.getMode().equals("client") || Conf.getMode().equals("both")) {
	    Viewer.setTextIntro("Starting GUI");

	    Debug.debugln(BddManager.class, "starting Viewer thread");
	    Viewer viewer = new Viewer();
 	    try {

		/*
		if (Conf.getMode().equals("client")) {
		    try {
			remoteAccess.Ping();
		    }
		    catch (RemoteException ex) {
			Debug.logExceptionAndExit(ex);
		    }
		    System.exit(0);
		}
		*/

		viewer.startThread(remoteAccess).join();
	    }
	    catch (InterruptedException ex) {
		Debug.logException(ex);
	    }
	}

	// try { Thread.currentThread().sleep(1000); } catch (InterruptedException ex) { Debug.logException(ex); }
    }
}

